import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path as PATH
from matplotlib.patches import Polygon
import matplotlib.patheffects as pe
import pandas as pd
import os,sys
import itertools
from Thermobar.core import *

def garnet_CARP_class_Griffin2002(gt_comps):

    '''
    A function to classify Cr-pyrope garnet xenocrysts according to
    the study of Griffin et al. (2002)
    :cite: `griffin2002`

    ###Parameters###
    gt_comps: garnet composition dataframe imported from core function
    '''

    idx_critical_missing = []
    for i in range(0,len(gt_comps)):

        if (pd.isna(gt_comps['CaO_Gt'][i]) == True) or (pd.isna(gt_comps['Cr2O3_Gt'][i]) == True) or\
         (pd.isna(gt_comps['MgO_Gt'][i]) == True) or (pd.isna(gt_comps['FeOt_Gt'][i]) == True) or\
         (pd.isna(gt_comps['TiO2_Gt'][i]) == True) or  (pd.isna(gt_comps['Zr_Gt'][i]) == True) or\
         (pd.isna(gt_comps['Y_Gt'][i]) == True):

            idx_critical_missing.append(i)

    if len(idx_critical_missing) > 0:
        print('WARNING')
        print('There are critical components missing in some of the samples for CARP analysis and might lead to wrong analysis. idx list: ')
        print(idx_critical_missing)

    h1 = []
    h2 = []
    h3 = []
    l1 = []
    l2 = []
    l3 = []
    l4 = []
    l5 = []
    l6 = []
    l7 = []
    l8 = []
    l9 = []
    l10a = []
    l10b = []
    l11 = []
    l12 = []
    l13 = []
    l14 = []
    l15 = []
    l16 = []
    l17 = []
    l18 = []
    l19 = []
    l20 = []
    l21 = []
    l22 = []
    l23 = []
    l24 = []
    l25 = []
    l26 = []
    l27 = []
    l28 = []
    carp_type_list = []

    #Defining CARP tree from Griffin et al. (2002, G-Cubed,doi:10.1029/2002gc000298)

    for i in range(0,len(np.array(gt_comps['CaO_Gt']))):

        if gt_comps['CaO_Gt'][i] < 4.0:
            if gt_comps['MgO_Gt'][i] < 21.05:
                h1.append(i)
                type = 'H1'
            else:
                if gt_comps['Y_Gt'][i] < 11.05:
                    h2.append(i)
                    type = 'H2'
                else:
                    h3.append(i)
                    type = 'H3'
        else:
            if gt_comps['MgO_Gt'][i] > 22.55:
                l28.append(i)
                type = 'L28'
            else:
                if gt_comps['Cr2O3_Gt'][i] < 0.73:
                    if gt_comps['FeOt_Gt'][i] < 9.1:
                        l1.append(i)
                        type = 'L1'
                    else:
                        l2.append(i)
                        type = 'L2'
                else:
                    if gt_comps['Zn_Gt'][i] > 22.0:
                        if gt_comps['TiO2_Gt'][i] < 0.47:
                            l25.append(i)
                            type = 'L25'
                        else:
                            if gt_comps['Zr_Gt'][i] < 35.0:
                                l26.append(i)
                                type = 'L26'
                            else:
                                l27.append(i)
                                type = 'L27'
                    else:
                        if gt_comps['Y_Gt'][i] < 5.0:
                            if gt_comps['TiO2_Gt'][i] > 0.135:
                                l7.append(i)
                                type = 'L7'
                            else:
                                if gt_comps['Zr_Gt'][i] < 14.5:
                                    l3.append(i)
                                    type = 'L3'
                                else:
                                    if gt_comps['Ga_Gt'][i] > 8.8:
                                        l6.append(i)
                                        type = 'L6'
                                    else:
                                        if gt_comps['Cr2O3_Gt'][i] < 3.8:
                                            l4.append(i)
                                            type = 'L4'
                                        else:
                                            l5.append(i)
                                            type = 'L5'
                        else:
                            if gt_comps['Cr2O3_Gt'][i] > 5.84:
                                if gt_comps['CaO_Gt'][i] < 5.4:
                                    l18.append(i)
                                    type = 'L18'
                                else:
                                    if gt_comps['FeOt_Gt'][i] > 8.3:
                                        l24.append(i)
                                        type = 'L24'
                                    else:
                                        if gt_comps['MgO_Gt'][i] > 20.45:
                                            if gt_comps['FeOt_Gt'][i] < 6.85:
                                                l22.append(i)
                                                type = 'L22'
                                            else:
                                                l23.append(i)
                                                type = 'L23'
                                        else:
                                            if gt_comps['Y_Gt'][i] < 15.4:
                                                l19.append(i)
                                                type = 'L19'
                                            else:
                                                if gt_comps['Zr_Gt'][i] < 53.7:
                                                    l20.append(i)
                                                    type = 'L20'
                                                else:
                                                    l21.append(i)
                                                    type = 'L21'
                            else:
                                if gt_comps['CaO_Gt'][i] > 6.09:
                                    l17.append(i)
                                    type = 'L17'
                                else:
                                    if gt_comps['Sr_Gt'][i] > 2.5:
                                        l16.append(i)
                                        type = 'L16'
                                    else:
                                        if gt_comps['MgO_Gt'][i] < 19.15:
                                            if gt_comps['FeOt_Gt'][i] < 8.1:
                                                l8.append(i)
                                                type = 'L8'
                                            else:
                                                l9.append(i)
                                                type = 'L9'
                                        else:
                                            if gt_comps['CaO_Gt'][i] > 5.54:
                                                if gt_comps['Cr2O3_Gt'][i] < 2.28:
                                                    l14.append(i)
                                                    type = 'L14'
                                                else:
                                                    l15.append(i)
                                                    type = 'L15'
                                            else:
                                                if gt_comps['Ga_Gt'][i] > 8.6:
                                                    if gt_comps['TiO2_Gt'][i] < 0.13:
                                                        l12.append(i)
                                                        type = 'L12'
                                                    else:
                                                        l13.append(i)
                                                        type = 'L13'
                                                else:
                                                    if gt_comps['TiO2_Gt'][i] > 0.37:
                                                        l11.append(i)
                                                        type = 'L11'
                                                    else:
                                                        if gt_comps['TiO2_Gt'][i] > 0.1:
                                                            l10b.append(i)
                                                            type = 'L10b'
                                                        else:
                                                            l10a.append(i)
                                                            type = 'L10a'
        carp_type_list.append(type)

    len_tot = len(h1) + len(h2) + len(h3) + len(l28) + len(l1) + len(l2) +\
    len(l25) + len(l26) + len(l27) + len(l7) + len(l3) + len(l6) +\
    len(l4) + len(l5) + len(l18) + len(l24) + len(l22) + len(l23) +\
    len(l19) + len(l20) + len(l21) + len(l17) + len(l16) + len(l8) +\
    len(l9) + len(l14) + len(l15) + len(l12) + len(l13) + len(l11) +\
    len(l10b) + len(l10a)
    carp_depleted_harz = h2
    carp_depleted_lherz = list(itertools.chain(l3,l5))
    carp_depleted_metasomatised = list(itertools.chain(h3,l15,
    l18,l19,l21))
    carp_fertile_lherz = list(itertools.chain(l9,l10a,l10b,l16))
    carp_melt_metas = list(itertools.chain(l13,l25,l27))
    carp_unclass = list(itertools.chain(h1,l4,l6,l7,l8,l22,
    l12,l14,l17,l20,l11,l23,l24,l26,l28))
    carp_depleted_harz.sort()
    carp_depleted_metasomatised.sort()
    carp_fertile_lherz.sort()
    carp_melt_metas.sort()
    carp_unclass.sort()

    return carp_depleted_harz, carp_depleted_lherz, carp_depleted_metasomatised, carp_fertile_lherz, carp_melt_metas, carp_unclass, len_tot

def garnet_class_Grutter2003(gt_comps):

    '''
    A function to classify Cr-pyrope garnet xenocrysts according to
    the study of Grutter et al. (2003)
    :cite: `grutter2003`

    ###Parameters###
    gt_comps: garnet composition dataframe imported from core function
    '''

    idx_critical_missing = []
    idx_na_missing = []
    for i in range(0,len(gt_comps)):
        if (pd.isna(gt_comps['CaO_Gt'][i]) == True) or (pd.isna(gt_comps['Cr2O3_Gt'][i]) == True) or\
         (pd.isna(gt_comps['MgO_Gt'][i]) == True) or (pd.isna(gt_comps['FeOt_Gt'][i]) == True) or\
         (pd.isna(gt_comps['MnO_Gt'][i]) == True) or  (pd.isna(gt_comps['TiO2_Gt'][i]) == True):

            idx_critical_missing.append(i)

        if (pd.isna(gt_comps['Na2O_Gt'][i]) == True):

            idx_na_missing.append(i)

    if len(idx_critical_missing) > 0:
        print('WARNING')
        print('There are critical components missing in some of the samples for Grutter2003 analysis. For these samples analyses might be wrong. idx list: ')
        print(idx_critical_missing)

    if len(idx_na_missing) > 0:
        print('WARNING')
        print('Na2O component is missing in some of the samples, diamondiferious classificatons might be wrong, defaulting to non-indicator classification. idx list: ')
        print(idx_na_missing)

    ca_int = []
    mg_num_g = []
    g10 = []
    g10d = []
    g9 = []
    g12 = []
    g1 = []
    g11 = []
    g5 = []
    g4 = []
    g4d = []
    g3 = []
    g3d = []
    g5d = []
    g_class = []
    g_all = []

    for i in range(0,len(np.array(gt_comps['CaO_Gt']))):
        if gt_comps['CaO_Gt'][i] < ((gt_comps['Cr2O3_Gt'][i]*0.25) + 3.375):
            ca_int.append((13.5*gt_comps['CaO_Gt'][i])/(gt_comps['Cr2O3_Gt'][i]+13.5))
        else:
            ca_int.append(gt_comps['CaO_Gt'][i] - (0.25*gt_comps['Cr2O3_Gt'][i]))

        mg_num_g.append((gt_comps['MgO_Gt'][i] / 40.3) / ((gt_comps['MgO_Gt'][i] / 40.3) + (gt_comps['FeOt_Gt'][i]/71.85)))

        stop_search = False
        one_before_g1 = False

        if (gt_comps['Cr2O3_Gt'][i] >= 1.0) and (gt_comps['Cr2O3_Gt'][i] < 22.0):
            if (ca_int[i] > 0.0) and (ca_int[i] < 3.375):
                if (mg_num_g[i] >= 0.75) and (mg_num_g[i] < 0.95):
                    if (gt_comps['Cr2O3_Gt'][i] >= ((0.94*gt_comps['CaO_Gt'][i]) + 5)) and (gt_comps['MnO_Gt'][i] < 0.36):
                        g10d.append(i)
                        g_class.append('G10D')
                        stop_search = True
                    else:
                        g10.append(i)
                        g_class.append('G10')
                        stop_search = True

        if stop_search == False:
            if (gt_comps['Cr2O3_Gt'][i] >= 1.0):
                if (ca_int[i] >= 3.375) and (ca_int[i] < 5.4):
                    if (mg_num_g[i] >= 0.70) and (mg_num_g[i] < 0.90):
                        g9.append(i)
                        g_class.append('G9')
                        stop_search = True
                    elif (mg_num_g[i] >= 0.3) and (mg_num_g[i] < 0.7):
                        if gt_comps['TiO2_Gt'][i] < (2.13 - (2.1 * mg_num_g[i])):
                            if (gt_comps['Na2O_Gt'][i] > 0.07):
                                g5d.append(i)
                                g_class.append('G5D')
                                stop_search = True
                            else:
                                g5.append(i)
                                g_class.append('G5')
                                stop_search = True


        if stop_search == False:
            if (gt_comps['TiO2_Gt'][i] >= (2.13 - 2.1*mg_num_g[i])) and (gt_comps['TiO2_Gt'][i] < 4.0):
                if (gt_comps['Cr2O3_Gt'][i] >= 0.0) and (gt_comps['Cr2O3_Gt'][i] < 4.0):
                    if (ca_int[i] >= 3.375) and (ca_int[i] < 6.0):
                        if (mg_num_g[i] >= 0.65) and (mg_num_g[i] < 0.85):
                            g1.append(i)
                            g_class.append('G1')
                            one_before_g1 = True


        if stop_search == False:
            if (gt_comps['TiO2_Gt'][i] >= (2.13 - 2.1*mg_num_g[i])) and (gt_comps['TiO2_Gt'][i] < 4.0):
                if (gt_comps['Cr2O3_Gt'][i] >= 1.0) and (gt_comps['Cr2O3_Gt'][i] < 20.0):
                    if (ca_int[i] >= 3.0):
                        if gt_comps['CaO_Gt'][i] < 28.0:
                            if (mg_num_g[i] >= 0.65) and (mg_num_g[i] < 0.90):
                                if one_before_g1 == False:
                                    g11.append(i)
                                    g_class.append('G11')
                                    stop_search = True
                                else:
                                    g11.append(i)
                                    del g1[-1]
                                    g_class[-1] == 'G11'
                                    stop_search = True

        if stop_search == False:
            if (gt_comps['Cr2O3_Gt'][i] >= 1.0) and (gt_comps['Cr2O3_Gt'][i] < 20.0):
                if (ca_int[i] >= 5.4):
                    if gt_comps['CaO_Gt'][i] < 28.0:
                        if gt_comps['MgO_Gt'][i] > 5.0:
                            if one_before_g1 == True:
                                pass
                            else:
                                g12.append(i)
                                g_class.append('G12')
                                stop_search = True

        if stop_search == False:
            if (gt_comps['Cr2O3_Gt'][i] < 1.0):
                if (gt_comps['CaO_Gt'][i] >= 2.0) and (gt_comps['CaO_Gt'][i] < 6.0):
                    if (mg_num_g[i] >= 0.3) and (mg_num_g[i] < 0.9):
                        if gt_comps['TiO2_Gt'][i] < (2.13 - (2.1 * mg_num_g[i])):
                            if (gt_comps['Na2O_Gt'][i] > 0.07):
                                g4d.append(i)
                                g_class.append('G4D')
                                stop_search = True
                            else:
                                g4.append(i)
                                g_class.append('G4')
                                stop_search = True

        if stop_search == False:
            if (gt_comps['Cr2O3_Gt'][i] >= 0.0) and (gt_comps['Cr2O3_Gt'][i] < 1.0):
                if (gt_comps['CaO_Gt'][i] >= 6.0) and (gt_comps['CaO_Gt'][i] < 32.0):
                    if (mg_num_g[i] >= 0.17) and (mg_num_g[i] < 0.86):
                        if gt_comps['TiO2_Gt'][i] < (2.13 - (2.1 * mg_num_g[i])):
                            if gt_comps['TiO2_Gt'][i] < 2.0:
                                if one_before_g1 == True:
                                    pass
                                else:
                                    if (gt_comps['Na2O_Gt'][i] > 0.07):
                                        g3d.append(i)
                                        g_class.append('G3D')
                                        stop_search = True
                                    else:
                                        g3.append(i)
                                        g_class.append('G3')
                                        stop_search = True

    total_G = list(itertools.chain(g10,g10d,g9,g12,
    g1,g11,g5,g5d,g4,g4d,g3,g3d))

    #Checking for duplicates
    total_G_duplicates = set([x for x in total_G if total_G.count(x) > 1])
    if len(total_G_duplicates) != 0:
        print('There are duplicates in the G-classification scheme, check whats wrong')
    else:
        len_tot_g = len(total_G)

    g0 = []
    for i in range(0,len(np.array(gt_comps['CaO_Gt']))):
        if (i in total_G == False):
            g0.append(i)

    return g_class

def garnet_ca_cr_class_Griffin2002(gt_comps):

    '''
    A function to make CaO-Cr2O3 classifications of Cr-pyrope garnet xenocrysts
    according to the study of Griffin et al. (2002)
    :cite: `griffin2002`

    ###Parameters###
    gt_comps: garnet composition dataframe imported from core function
    '''

    #Checking if any CaO or Cr2O3 is missing

    for i in range(0,len(gt_comps)):

        if (pd.isna(gt_comps['CaO_Gt'][i]) == True) or (pd.isna(gt_comps['Cr2O3_Gt'][i]) == True):

            print('WARNING')
            print('There is CaO or Cr2O3 missing in one of the samples. idx = ' + str(i))

    low_cr_ca = [0,9,9,0]
    low_cr_cr = [0,0,1.52,1.52]

    low_cr_polygon = []
    for i in range(0,len(low_cr_ca)):
        low_cr_polygon.append((low_cr_ca[i],low_cr_cr[i]))
    low_cr_polygon_shp = PATH(low_cr_polygon)

    low_ca_harz_ca = [2.17,5.21,0.06832,0.029933,0.00055432]
    low_ca_harz_cr = [1.5201,12.01,12.01,6.9085,1.5305]

    low_ca_harz_polygon = []
    for i in range(0,len(low_ca_harz_ca)):
        low_ca_harz_polygon.append((low_ca_harz_ca[i],low_ca_harz_cr[i]))
    low_ca_harz_polygon_shp = PATH(low_ca_harz_polygon)

    ca_harz_ca = [2.172,5.215,7.062,3.914]
    ca_harz_cr = [1.52,12.001,12.001,1.52]

    ca_harz_polygon = []
    for i in range(0,len(ca_harz_cr)):
        ca_harz_polygon.append((ca_harz_ca[i],ca_harz_cr[i]))
    ca_harz_polygon_shp = PATH(ca_harz_polygon)

    lherz_ca = [3.9141,7.06205,8.73,5.75]
    lherz_cr = [1.5201,12.001,12.001,1.5201]

    lherz_polygon = []
    for i in range(0,len(ca_harz_cr)):
        lherz_polygon.append((lherz_ca[i],lherz_cr[i]))
    lherz_polygon_shp = PATH(lherz_polygon)

    wehrl_ca = [5.752,8.74,9,9]
    wehrl_cr = [1.5201,12.01,12.01,1.5201]

    wehrl_polygon = []
    for i in range(0,len(wehrl_cr)):
        wehrl_polygon.append((wehrl_ca[i],wehrl_cr[i]))
    wehrl_polygon_shp = PATH(wehrl_polygon)

    cacr_class = []
    list_cacr_class_list = []
    for i in range(0,len(np.array(gt_comps['CaO_Gt']))):
        list_cacr_class_list.append((gt_comps['CaO_Gt'][i],gt_comps['Cr2O3_Gt'][i]))
    a = low_cr_polygon_shp.contains_points(list_cacr_class_list)
    b = low_ca_harz_polygon_shp.contains_points(list_cacr_class_list)
    c = ca_harz_polygon_shp.contains_points(list_cacr_class_list)
    d = lherz_polygon_shp.contains_points(list_cacr_class_list)
    e = wehrl_polygon_shp.contains_points(list_cacr_class_list)
    for i in range(0,len(list_cacr_class_list)):
        check = False
        if a[i] == True:
            cacr_class.append('Low-Cr Peridotite')
            check = True
        if b[i] == True:
            cacr_class.append('Low-Ca Harzburgite')
            check = True
        if c[i] == True:
            cacr_class.append('Ca-Harzburgite')
            check = True
        if d[i] == True:
            cacr_class.append('Lherzolite')
            check = True
        if e[i] == True:
            cacr_class.append('Wehrlite')
            check == True
        if check == False:
            cacr_class.append('Unclassified')

    return cacr_class

def y_zr_classification_Griffin2002(gt_comps):

    '''
    A function to make Y-Zr classifications of Cr-pyrope garnet xenocrysts
    according to the study of Griffin et al. (2002)
    :cite: `griffin2002`

    ###Parameters###
    gt_comps: garnet composition dataframe imported from core function
    '''

    Zr_depleted = [0.1003,3.634,7.168,10.14,13.49,15.91,17.96,19.81,
    21.67,23.34,24.45,25.38,26.49,27.41,28.15,28.88,29.25,29.98,29.98,
    30.16,30.34,30.33,30.51,0.5567]
    Y_depleted = [13.59,13.4,13.34,13.21,13.08,13.08,12.89,12.55,
    12.15,11.35,10.75,9.948,9.213,8.343,7.339,6.536,5.397,4.326,3.455,
    2.517,1.579,0.3723,-0.02949,-0.07968]
    depleted_polygon = []
    for i in range(0,len(Zr_depleted)):
        depleted_polygon.append((Zr_depleted[i],Y_depleted[i]))
    depleted_polygon_shp = PATH(depleted_polygon)


    Zr_melt = [16.656,18.149,20.389,22.813,24.492,27.102,31.394,35.497,39.787,
    44.079,49.299,55.824,61.603,67.381,73.719,80.055,86.576,92.914,99.06,105.58,113.22,
    119.36,126.99,134.63,141.33,141.1,135.88,131.77,127.29,123,117.58,111.99,105.84,
    100.43,94.093,87.568,82.164,76.015,69.121,62.975,55.526,47.893,41.19,33.745,27.787,
    26.492,23.714,21.302,18.886]
    Y_melt = [13.218,13.958,14.833,15.574,16.247,16.989,18.672,20.019,21.434,23.184,
    24.667,26.42,27.972,29.456,31.075,32.493,33.778,35.464,36.278,37.362,
    38.447,39.261,40.012,40.695,41.309,35.343,33.927,32.245,29.958,27.739,25.586,23.633,
    22.081,20.463,18.844,17.023,15.942,14.658,13.306,12.357,11.474,10.791,9.975,9.5604,
    9.0142,10.017,11.688,12.556,12.887]
    melt_polygon = []
    for i in range(0,len(Zr_melt)):
        melt_polygon.append((Zr_melt[i],Y_melt[i]))
    melt_polygon_shp = PATH(melt_polygon)

    Zr_fertile = [0.35165,5.561,11.145,17.107,21.953,27.175,31.468,
    35.203,38.382,40.819,42.887,44.21,45.163,45.373,45.207,45.807,
    -0.14714,-10.465,-9.482]
    Y_fertile = [22.373,22.382,22.726,23.876,25.023,26.775,28.658,
    30.541,32.691,35.175,38.128,40.878,43.896,47.113,49.861,55.425,
    55.348,43.937,25.976]
    fertile_polygon = []
    for i in range(0,len(Zr_fertile)):
        fertile_polygon.append((Zr_fertile[i],Y_fertile[i]))
    fertile_polygon_shp = PATH(fertile_polygon)

    Zr_phlg = [28.34,30.918,35.569,45.058,56.967,67.198,80.409,89.713,
    96.79,107.04,113.76,119.36,125.15,130.56,137.84,140.46,139.76,135.85,130.62,122.97,
    115.7,108.24,100.97,93.696,85.683,77.11,66.865,50.856,42.297,35.598]
    Y_phlg = [8.2779,4.7299,4.6707,4.7536,5.0416,4.8577,5.0139,5.3646,
    6.2478,8.8789,11.973,14.731,17.086,19.374,22.536,24.216,30.918,29.303,27.216,
    24.522,22.097,19.873,17.649,15.492,13.803,11.711,10.085,8.9188,8.7704,8.5581]
    phlg_polygon = []
    for i in range(0,len(Zr_phlg)):
        phlg_polygon.append((Zr_phlg[i],Y_phlg[i]))
    phlg_polygon_shp = PATH(phlg_polygon)

    #Classification for lists
    yzr_class = []
    list_yzr_class_list = []
    for i in range(0,len(np.array(gt_comps['Y_Gt']))):
        list_yzr_class_list.append((gt_comps['Zr_Gt'][i],gt_comps['Y_Gt'][i]))
    a = melt_polygon_shp.contains_points(list_yzr_class_list)
    b = phlg_polygon_shp.contains_points(list_yzr_class_list)
    c = fertile_polygon_shp.contains_points(list_yzr_class_list)
    d = depleted_polygon_shp.contains_points(list_yzr_class_list)
    for i in range(0,len(list_yzr_class_list)):
        check = False
        if a[i] == True:
            yzr_class.append('Melt Metasomatised')
            check = True
        if b[i] == True:
            yzr_class.append('Phlogopite Metasomatism')
            check = True
        if c[i] == True:
            yzr_class.append('Fertile')
            check = True
        if d[i] == True:
            yzr_class.append('Depleted')
            check = True
        if check == False:
            yzr_class.append('Unclassified')

    return yzr_class

def calculate_ol_mg(gt_comps, T_Ni, xMg, xCa, xFe):

    '''
    A function to derive olivine Mg# from coxisting Cr-pyrope garnet xenocrysts.
    :cite: `gaul2000`

    ###Parameters###
    gt_comps: garnet composition dataframe imported from core function
    T_Ni: Temperature output taken from one of the garnet thermometry methods in
    Kelvin.

    Returns
    --------------
    Numpy array
        Ol Mg#
    '''

    mg_ol = np.zeros(len(gt_comps['MgO_Gt']))

    T_Ni = np.array(T_Ni) - 273.0 #converting to celsius for calculations

    for i in range(0,len(np.array(gt_comps['MgO_Gt']))):

        p_calc = (0.00001786*(T_Ni[i]**2.0)) + (0.027419*T_Ni[i]) - 1.198
        p1 = p_calc - (0.000263*(p_calc**2.0)) - 29.76
        p2 = p_calc - (0.00039*(p_calc**2.0)) - 29.65
        p3 = p_calc - (0.000236*(p_calc**2.0)) - 29.79
        p4 = p_calc - (0.00045*(p_calc**2.0)) - 29.6
        tk = T_Ni[i] + 273.0
        dv1 = -462.5*(1.0191+((tk-1073)*0.0000287))*p1
        dv2 = -262.4*(1.0292+((tk-1073)*0.000045))*p2
        dv3 = 454.0*(1.02+((tk-1073)*0.0000284))*p3
        dv4 = 278.3*(1.0234+((tk-1073)*0.000023))*p4
        #(1347*V2+902+AJ2+(0.9-(1-0.9))*(498+1.51*(R2-30))-98*(T2-U2))/W2-0.357
        dv = dv1 + dv2 + dv3 + dv4
        kd = np.exp((((1347*xCa[i])+902+dv+(0.9-(1-0.9))*(498+1.51*(p_calc-30))-98*(xMg[i]-xFe[i]))/tk) - 0.357)
        xmgfe = kd * (xMg[i]/xFe[i])
        mg_ol[i] = xmgfe / (1+xmgfe)

    return mg_ol

def calculate_al2o3_whole_rock(gt_comps):

    '''
    A function to derive whole-rock Al2O3 from coxisting Cr-pyrope garnets xenocrysts.
    :cite: `oreilly2006`

    ###Parameters###
    gt_comps: garnet composition dataframe imported from core function

    Returns
    --------------
    Numpy array
        Whole rock Al2O3 values.

    '''

    wr_al_y = np.zeros(len(gt_comps['Y_Gt']))

    for i in range(0,len(gt_comps['Y_Gt'])):

        if (gt_comps['Y_Gt'][i] > 0.0):
            wr_al_y[i] = (10.0**(0.778-0.14*(-3.2253*np.log(gt_comps['Y_Gt'][i])+13.006)))
        else:
            wr_al_y[i] = None

    return wr_al_y



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd


def rasmussen(P_kbar):
    """ Linear fit to the supporting information of Rasmussen et al. 2022,
    overall best fit density vs. depth"""
    P=P_kbar
    D=3.784444149179223*P+0.2857169093579655
    return D

def hill_zucca(P_kbar):
    """ Parameterization of Hill and Zucca (1987),
    as given in Putirka (2017) Down the Crater Elements supplement
    """
    P=P_kbar

    D=-2.77*10**(-5) * (P**4) - 2.0*10**(-3)*(P**3) - 4.88*10**(-2)*P**2 + 3.6*P - 6.34*(10**(-2))

    return D

def ryan_lerner(P_kbar):
    """ Parameterization of Ryan 1987, actual equation from Lerner et al. 2021
    After 16.88 km (455 MPa), assume density is 2.746, as density turns around again. Used for Hawaii
    """
    P=P_kbar*100
    if P<455.09090909:
        D=(4.578*10**(-8) *P**3) - (4.151*10**(-5) *P**2) + (4.652*10**(-2) *P)
    else:
        D=P/(9.8*2.749643038642074)

    return D

def mavko_debari(P_kbar):
    """ Parameterization of Mavko and Thompson (1983) and DeBari and Greene (2011)
    as given in Putirka (2017) Down the Crater Elements supplement, used for Cascades
    """
    P=P_kbar
    D=0.4853881 + 3.6006116*P - 0.0117368*(P-1.3822)**2


    return D


def prezzi(P_kbar):
    """ Parameterization of Prezzi et al. (2009),
    as given in Putirka (2017) Down the Crater Elements supplement.
    Used for Andes.
    """
    P=P_kbar
    D=4.88 + 3.30*P - 0.0137*(P - 18.01)**2

    return D


Profile_funcs={ryan_lerner, mavko_debari, hill_zucca, prezzi, rasmussen}
Profile_funcs_by_name= {p.__name__: p for p in Profile_funcs}

def convert_pressure_to_depth(P_kbar=None, model=None, crust_dens_kgm3=None):
    """ Converts pressure in kbar to depth in km using a variety of crustal density profiles


    Parameters
    -----------

    P_kbar: int, float, pd.Series, np.ndarray
        Pressure in kbar

    model: str or float:
    If you want an existing model, choose from
        ryan_lerner: Parameterization of Ryan 1987, actual equation from Lerner et al. 2021
    After 16.88 km (455 MPa), assume density is 2.746, as density turns around again. This profile is tweaked for Hawaii

        mavko_debari: Parameterization of Mavko and Thompson (1983) and DeBari and Greene (2011)
    as given in Putirka (2017) Down the Crater Elements supplement.
    **Currently has a typo, have emailed Keith Putirka!***

        hill_zucca: Parameterization of Hill and Zucca (1987),
    as given in Putirka (2017) Down the Crater Elements supplement

        prezzi: Parameterization of Prezzi et al. (2009),
    as given in Putirka (2017) Down the Crater Elements supplement. Tweaked for Andes.

        rasmussen: Linear fit to the supporting information of Rasmussen et al. 2022,
    overall best fit density vs. depth

    OR

    crust_dens_kgm3: float
        Crustal density in kg/m3

    Else, just enter a crustal density in kg/m3, e.g., model=2700



    Returns
    -----------

    Depth in km as a panda series

    """
    if model is not None:

        try:
            func = Profile_funcs_by_name[model]
        except KeyError:
            raise ValueError(f'{model} is not a valid model') from None
        #sig=inspect.signature(func)
        if isinstance(P_kbar, float) or isinstance(P_kbar, int):
            D=func(P_kbar)

        if isinstance(P_kbar, pd.Series):
            D=np.empty(len(P_kbar), float)
            for i in range(0, len(P_kbar)):
                D[i]=func(P_kbar.iloc[i])

        if isinstance(P_kbar, np.ndarray):
            D=np.empty(len(P_kbar), float)
            for i in range(0, len(P_kbar)):
                D[i]=func(P_kbar[i])



    if crust_dens_kgm3 is not None:
        D=10**5*P_kbar/(9.8*crust_dens_kgm3)

    D_series=pd.Series(D)
    return D_series



import numpy as np
import matplotlib.pyplot as plt
from functools import partial
import inspect
import warnings as w
import numbers
import pandas as pd



from Thermobar.core import *

## Pressure equations for two pyroxenes


def P_Put2008_eq38(T=None, *, Na_Opx_cat_6ox, Al_IV_Opx_cat_6ox, Al_VI_Opx_cat_6ox,
Ti_Opx_cat_6ox, Ca_Opx_cat_6ox, Cr_Opx_cat_6ox, Mg_Opx_cat_6ox,
Fet_Opx_cat_6ox, Mn_Opx_cat_6ox, Ca_Cpx_cat_6ox, Fm2Si2O6, En_Opx, Di_Opx):
    '''

    Two pyroxene barometer of Putirka (2008) Eq38. Calibrated on Mg#-rich systems (>0.75)
    :cite:`putirka2008thermometers`

    | SEE=+-3.7 kbar
    '''
    Lindley_Fe3_Opx = (Na_Opx_cat_6ox + Al_IV_Opx_cat_6ox - Al_VI_Opx_cat_6ox - \
        2 * Ti_Opx_cat_6ox - Cr_Opx_cat_6ox)  # This is cell FR
    Lindley_Fe3_Opx[Lindley_Fe3_Opx < 0] = 0
    a_En_opx_mod = (((0.5 * Mg_Opx_cat_6ox / (0.5 * (Fet_Opx_cat_6ox - Lindley_Fe3_Opx)
    + 0.5 * Mg_Opx_cat_6ox + Na_Opx_cat_6ox +Ca_Opx_cat_6ox + Mn_Opx_cat_6ox)))
    * (0.5 * Mg_Opx_cat_6ox / (0.5 * Mg_Opx_cat_6ox + 0.5 * (Fet_Opx_cat_6ox - Lindley_Fe3_Opx)
    + Ti_Opx_cat_6ox + Al_VI_Opx_cat_6ox + Cr_Opx_cat_6ox + Lindley_Fe3_Opx)))
    Kf = Ca_Opx_cat_6ox / (1 - Ca_Cpx_cat_6ox)
    return (-279.8 + 293 * Al_VI_Opx_cat_6ox + 455 * Na_Opx_cat_6ox + 229 * Cr_Opx_cat_6ox +
            519 * Fm2Si2O6 - 563 * En_Opx + 371 * Di_Opx + 327 * a_En_opx_mod + 1.19 / Kf)


def P_Put2008_eq39(T, *, Na_Opx_cat_6ox, Al_IV_Opx_cat_6ox, Al_VI_Opx_cat_6ox,
Ti_Opx_cat_6ox, Cr_Opx_cat_6ox, Fet_Opx_cat_6ox, Mn_Opx_cat_6ox, Ca_Opx_cat_6ox,
Mg_Opx_cat_6ox, Na_Cpx_cat_6ox, Al_IV_cat_6ox, Al_VI_cat_6ox, Ti_Cpx_cat_6ox,
Ca_Cpx_cat_6ox, Mg_Cpx_cat_6ox, Mn_Cpx_cat_6ox, Fet_Cpx_cat_6ox, Cr_Cpx_cat_6ox,
Fm2Si2O6, En_Opx, EnFs):
    '''

    Two pyroxene barometer of Putirka (2008) Eq39. Similar to Eq38, but
    has a temperature term.
    :cite:`putirka2008thermometers`

    | SEE=+-2.8 kbar (Cpx Mg#>0.75)
    | SEE=+-3.2 kbar (all data)
    '''
    Lindley_Fe3_Opx = Na_Opx_cat_6ox + Al_IV_Opx_cat_6ox - \
        Al_VI_Opx_cat_6ox - 2 * Ti_Opx_cat_6ox - Cr_Opx_cat_6ox
    Lindley_Fe3_Opx[Lindley_Fe3_Opx < 0] = 0
    a_En_opx_mod = (((0.5 * Mg_Opx_cat_6ox / (0.5 * (Fet_Opx_cat_6ox - Lindley_Fe3_Opx)
    + 0.5 * Mg_Opx_cat_6ox + Na_Opx_cat_6ox +Ca_Opx_cat_6ox + Mn_Opx_cat_6ox)))
    * (0.5 * Mg_Opx_cat_6ox / (0.5 * Mg_Opx_cat_6ox + 0.5 * (Fet_Opx_cat_6ox - Lindley_Fe3_Opx)
    + Ti_Opx_cat_6ox + Al_VI_Opx_cat_6ox + Cr_Opx_cat_6ox + Lindley_Fe3_Opx)))

    Lindley_Fe3_Cpx = Na_Cpx_cat_6ox + Al_IV_cat_6ox - \
        Al_VI_cat_6ox - 2 * Ti_Cpx_cat_6ox - Cr_Cpx_cat_6ox
    Lindley_Fe3_Cpx[Lindley_Fe3_Cpx < 0] = 0
    a_Di_cpx = Ca_Cpx_cat_6ox / (Ca_Cpx_cat_6ox + 0.5 * Mg_Cpx_cat_6ox + 0.5 * (
        Fet_Cpx_cat_6ox - Lindley_Fe3_Cpx) + Mn_Cpx_cat_6ox + Na_Cpx_cat_6ox)
    Kf = Ca_Opx_cat_6ox / (1 - Ca_Cpx_cat_6ox)
    return (-94.25 + 0.045 * (T - 273.15) + 187.7 * Al_VI_Opx_cat_6ox + 246.8 * Fm2Si2O6 -
            212.5 * En_Opx + 127.5 * a_En_opx_mod - 69.4 * EnFs - 133.9 * a_Di_cpx - 1.66 / Kf)

## Temperature equations for two pyroxenes


def T_Put2008_eq36(P, *, EnFs, Fm2Si2O6, Ca_Cpx_cat_6ox,
                   CrCaTs, Mn_Opx_cat_6ox, Na_Opx_cat_6ox, En_Opx, Di_Opx):
    '''
    Two-pyroxene thermometer of Putirka (2008) eq 36. Best for Cpx with Mg#>0.75
    :cite:`putirka2008thermometers`

    SEE=+-45C for Cpx Mg#>0.75
    SEE=+-56C for all data
    '''
    return (273.15 + 10 ** 4 / (11.2 - 1.96 * np.log(EnFs.astype(float) / Fm2Si2O6.astype(float)) - 3.3 * Ca_Cpx_cat_6ox - 25.8 *
            CrCaTs + 33.2 * Mn_Opx_cat_6ox - 23.6 * Na_Opx_cat_6ox - 2.08 * En_Opx - 8.33 * Di_Opx - 0.05 * P))


def T_Brey1990(P, *, Fet_Cpx_cat_6ox, Ca_Cpx_cat_6ox, Mg_Cpx_cat_6ox, Na_Cpx_cat_6ox,
               Fet_Opx_cat_6ox, Mg_Opx_cat_6ox, Ca_Opx_cat_6ox, Na_Opx_cat_6ox):
    '''
    Two-pyroxene thermometer of Brey and Kohler (1990).
    :cite:`brey1990geothermobarometry`

    SEE=+-50C for Cpx Mg#>0.75
    SEE=+-70C for all data
    '''
    return ((23664 + (24.9 + 126.3 * Fet_Cpx_cat_6ox / (Fet_Cpx_cat_6ox + Mg_Cpx_cat_6ox)) * P)
    / (13.38 + (np.log((1 - Ca_Cpx_cat_6ox.astype(float) /(1 - Na_Cpx_cat_6ox.astype(float))) /
    (1 - Ca_Opx_cat_6ox.astype(float) / (1 - Na_Opx_cat_6ox.astype(float)))))**2
    + 11.59 * Fet_Opx_cat_6ox / (Fet_Opx_cat_6ox + Mg_Opx_cat_6ox)))


def T_Put2008_eq37(P, *, EnFs, Di_Cpx, Fm2Si2O6, Mn_Opx_cat_6ox,
                   FmAl2SiO6, Mg_Cpx_cat_6ox, Fet_Cpx_cat_6ox):
    '''
    Two-pyroxene thermometer of Putirka (2008) eq 37. For Cpx with Mg#<0.75
    :cite:`putirka2008thermometers`

    SEE=+-38C for Cpx Mg#>0.75
    SEE=+-60C for all data
    '''
    return (273.15 + 10**4 / (13.4 - 3.4 * np.log(EnFs.astype(float) / Fm2Si2O6.astype(float)) + 5.59 * np.log(Mg_Cpx_cat_6ox.astype(float))
    + 23.85 * Mn_Opx_cat_6ox +6.48 * FmAl2SiO6 - 2.38 * Di_Cpx - 0.044 * P
    - 8.8 * Mg_Cpx_cat_6ox / (Mg_Cpx_cat_6ox + Fet_Cpx_cat_6ox)))


def T_Wood1973(P=None, *, Mg_Opx_cat_6ox, Ca_Opx_cat_6ox, Mn_Opx_cat_6ox,
Fet_Opx_cat_6ox, Na_Opx_cat_6ox, Al_IV_Opx_cat_6ox, Al_VI_Opx_cat_6ox,
Ti_Opx_cat_6ox, Cr_Opx_cat_6ox, Mg_Cpx_cat_6ox, Ca_Cpx_cat_6ox,
Mn_Cpx_cat_6ox, Fet_Cpx_cat_6ox, Na_Cpx_cat_6ox, Al_IV_cat_6ox,
Al_VI_cat_6ox, Ti_Cpx_cat_6ox, Cr_Cpx_cat_6ox):
    '''
    Two-pyroxene thermometer of Wood and Banno (1973)
    :cite:`wood1973garnet`

    '''
    # Opx parts
    Lindley_Fe3_Opx = Na_Opx_cat_6ox + Al_IV_Opx_cat_6ox - Al_VI_Opx_cat_6ox - \
        2 * Ti_Opx_cat_6ox - Cr_Opx_cat_6ox  # This is cell FR
    Lindley_Fe3_Opx[Lindley_Fe3_Opx < 0] = 0
    MgNo_WB_Opx = Mg_Opx_cat_6ox / \
        (Mg_Opx_cat_6ox + (Fet_Opx_cat_6ox - Lindley_Fe3_Opx))
    X_Mg_M2_Opx = (1 - Ca_Opx_cat_6ox - Na_Opx_cat_6ox -
                   Mn_Opx_cat_6ox) * MgNo_WB_Opx  # FL
    X_Fe_M2_Opx = (1 - Ca_Opx_cat_6ox - Na_Opx_cat_6ox -
                   Mn_Opx_cat_6ox) * (1 - MgNo_WB_Opx)  # FM
    X_Mg_M1_Opx = (1 - Lindley_Fe3_Opx - Al_VI_Opx_cat_6ox -
                   Ti_Opx_cat_6ox - Cr_Opx_cat_6ox) * MgNo_WB_Opx  # FJ
    X_Fe_M1_Opx = (1 - Lindley_Fe3_Opx - Al_VI_Opx_cat_6ox -
                   Ti_Opx_cat_6ox - Cr_Opx_cat_6ox) * (1 - MgNo_WB_Opx)  # FK
    MgNo_WB_Opx = Mg_Opx_cat_6ox / \
        (Mg_Opx_cat_6ox + (Fet_Opx_cat_6ox - Lindley_Fe3_Opx))
    a_opx_En = (X_Mg_M2_Opx / (X_Mg_M2_Opx + X_Fe_M2_Opx + Ca_Opx_cat_6ox +
    Na_Opx_cat_6ox + Mn_Opx_cat_6ox)) * \
        (X_Mg_M1_Opx / (Lindley_Fe3_Opx + Ti_Opx_cat_6ox +
         Al_VI_Opx_cat_6ox + Cr_Opx_cat_6ox + X_Mg_M1_Opx + X_Fe_M1_Opx))
    # Cpx parts
    Lindley_Fe3_Cpx = Na_Cpx_cat_6ox + Al_IV_cat_6ox - Al_VI_cat_6ox - \
        2 * Ti_Cpx_cat_6ox - Cr_Cpx_cat_6ox  # This is cell FR
    Lindley_Fe3_Cpx[Lindley_Fe3_Cpx < 0] = 0
    Fe2_WB_Cpx = Fet_Cpx_cat_6ox - Lindley_Fe3_Cpx
    MgNo_WB_Cpx = Mg_Cpx_cat_6ox / (Mg_Cpx_cat_6ox + Fe2_WB_Cpx)
    X_Mg_M2_Cpx = (1 - Ca_Cpx_cat_6ox - Na_Cpx_cat_6ox -
                   Mn_Cpx_cat_6ox) * MgNo_WB_Cpx  # FL
    X_Fe_M2_Cpx = (1 - Ca_Cpx_cat_6ox - Na_Cpx_cat_6ox -
                   Mn_Cpx_cat_6ox) * (1 - MgNo_WB_Cpx)  # FM
    X_Mg_M1_Cpx = (1 - Lindley_Fe3_Cpx - Al_VI_cat_6ox -
                   Ti_Cpx_cat_6ox - Cr_Cpx_cat_6ox) * MgNo_WB_Cpx  # FJ
    X_Fe_M1_Cpx = (1 - Lindley_Fe3_Cpx - Al_VI_cat_6ox -
                   Ti_Cpx_cat_6ox - Cr_Cpx_cat_6ox) * (1 - MgNo_WB_Cpx)  # FK
    a_cpx_En = (X_Mg_M2_Cpx / (X_Mg_M2_Cpx + X_Fe_M2_Cpx + Ca_Cpx_cat_6ox
    + Na_Cpx_cat_6ox + Mn_Cpx_cat_6ox)) * \
        (X_Mg_M1_Cpx / (Lindley_Fe3_Cpx + Ti_Cpx_cat_6ox +
         Al_VI_cat_6ox + Cr_Cpx_cat_6ox + X_Mg_M1_Cpx + X_Fe_M1_Cpx))

    return ((-10202 / (np.log(a_cpx_En.astype(float) / a_opx_En.astype(float)) - 7.65 *
            (1 - MgNo_WB_Opx) + 3.88 * (1 - MgNo_WB_Opx)**2 - 4.6)))


def T_Wells1977(P=None, *, Mg_Opx_cat_6ox, Ca_Opx_cat_6ox, Mn_Opx_cat_6ox,
Fet_Opx_cat_6ox, Na_Opx_cat_6ox, Al_IV_Opx_cat_6ox, Al_VI_Opx_cat_6ox,
 Ti_Opx_cat_6ox, Cr_Opx_cat_6ox, Mg_Cpx_cat_6ox, Ca_Cpx_cat_6ox,
 Mn_Cpx_cat_6ox, Fet_Cpx_cat_6ox, Na_Cpx_cat_6ox, Al_IV_cat_6ox,
 Al_VI_cat_6ox, Ti_Cpx_cat_6ox, Cr_Cpx_cat_6ox):
    '''
    Two-pyroxene thermometer of Wells 1977
    :cite:`wells1977pyroxene`

    '''
    # Opx parts
    Lindley_Fe3_Opx = Na_Opx_cat_6ox + Al_IV_Opx_cat_6ox - Al_VI_Opx_cat_6ox - \
        2 * Ti_Opx_cat_6ox - Cr_Opx_cat_6ox  # This is cell FR
    Lindley_Fe3_Opx[Lindley_Fe3_Opx < 0] = 0
    MgNo_WB_Opx = Mg_Opx_cat_6ox / \
        (Mg_Opx_cat_6ox + (Fet_Opx_cat_6ox - Lindley_Fe3_Opx))
    X_Mg_M2_Opx = (1 - Ca_Opx_cat_6ox - Na_Opx_cat_6ox -
                   Mn_Opx_cat_6ox) * MgNo_WB_Opx  # FL
    X_Fe_M2_Opx = (1 - Ca_Opx_cat_6ox - Na_Opx_cat_6ox -
                   Mn_Opx_cat_6ox) * (1 - MgNo_WB_Opx)  # FM
    X_Mg_M1_Opx = (1 - Lindley_Fe3_Opx - Al_VI_Opx_cat_6ox -
                   Ti_Opx_cat_6ox - Cr_Opx_cat_6ox) * MgNo_WB_Opx  # FJ
    X_Fe_M1_Opx = (1 - Lindley_Fe3_Opx - Al_VI_Opx_cat_6ox -
                   Ti_Opx_cat_6ox - Cr_Opx_cat_6ox) * (1 - MgNo_WB_Opx)  # FK
    MgNo_WB_Opx = Mg_Opx_cat_6ox / \
        (Mg_Opx_cat_6ox + (Fet_Opx_cat_6ox - Lindley_Fe3_Opx))
    a_opx_En = (X_Mg_M2_Opx / (X_Mg_M2_Opx + X_Fe_M2_Opx + Ca_Opx_cat_6ox
    + Na_Opx_cat_6ox + Mn_Opx_cat_6ox)) * \
        (X_Mg_M1_Opx / (Lindley_Fe3_Opx + Ti_Opx_cat_6ox +
         Al_VI_Opx_cat_6ox + Cr_Opx_cat_6ox + X_Mg_M1_Opx + X_Fe_M1_Opx))
    # Cpx parts
    Lindley_Fe3_Cpx = Na_Cpx_cat_6ox + Al_IV_cat_6ox - Al_VI_cat_6ox - \
        2 * Ti_Cpx_cat_6ox - Cr_Cpx_cat_6ox  # This is cell FR
    Lindley_Fe3_Cpx[Lindley_Fe3_Cpx < 0] = 0
    Fe2_WB_Cpx = Fet_Cpx_cat_6ox - Lindley_Fe3_Cpx
    MgNo_WB_Cpx = Mg_Cpx_cat_6ox / (Mg_Cpx_cat_6ox + Fe2_WB_Cpx)
    X_Mg_M2_Cpx = (1 - Ca_Cpx_cat_6ox - Na_Cpx_cat_6ox -
                   Mn_Cpx_cat_6ox) * MgNo_WB_Cpx  # FL
    X_Fe_M2_Cpx = (1 - Ca_Cpx_cat_6ox - Na_Cpx_cat_6ox -
                   Mn_Cpx_cat_6ox) * (1 - MgNo_WB_Cpx)  # FM
    X_Mg_M1_Cpx = (1 - Lindley_Fe3_Cpx - Al_VI_cat_6ox -
                   Ti_Cpx_cat_6ox - Cr_Cpx_cat_6ox) * MgNo_WB_Cpx  # FJ
    X_Fe_M1_Cpx = (1 - Lindley_Fe3_Cpx - Al_VI_cat_6ox -
                   Ti_Cpx_cat_6ox - Cr_Cpx_cat_6ox) * (1 - MgNo_WB_Cpx)  # FK
    a_cpx_En = (X_Mg_M2_Cpx / (X_Mg_M2_Cpx + X_Fe_M2_Cpx + Ca_Cpx_cat_6ox + Na_Cpx_cat_6ox + Mn_Cpx_cat_6ox)) * \
        (X_Mg_M1_Cpx / (Lindley_Fe3_Cpx + Ti_Cpx_cat_6ox +
         Al_VI_cat_6ox + Cr_Cpx_cat_6ox + X_Mg_M1_Cpx + X_Fe_M1_Cpx))

    return ((7341 / (3.355 + 2.44 * (1 - MgNo_WB_Opx) - np.log(a_cpx_En.astype(float) / a_opx_En.astype(float)))))

## Function for calculating Cpx-Opx pressure

Cpx_Opx_P_funcs = {P_Put2008_eq38, P_Put2008_eq39} # put on outside

Cpx_Opx_P_funcs_by_name = {p.__name__: p for p in Cpx_Opx_P_funcs}

def calculate_cpx_opx_press(*, cpx_comps=None, opx_comps=None,
Two_Px_Match=None, equationP=None, eq_tests=False, T=None):
    '''
    calculates pressure in kbar for Opx-Cpx pairs

    The function requires inputs of cpx_comps and opx_comps, or input of a
    combined dataframe of cpx-opx compositions (this is used for the
    calculate_cpx_opx_press_temp_matching function)

    Parameters
    -----------

    cpx_comps: pandas.DataFrame
        Clinopyroxene compositions with column headings SiO2_Cpx, MgO_Cpx etc.

    opx_comps: pandas.DataFrame
       Opx compositions with column headings SiO2_Opx, MgO_Opx etc.

    Two_Px_Match: pandas.DataFrame
        Combined Cpx-Opx compositions instead of separate dataframes.
        Used for calculate Cpx_Opx_press_temp_matching function.

    equationP: str
        Choose from:

        |  P_Put2008_eq38 (T-independent)
        |  P_Put2008_eq39 (T-dependent)

    T: float, int, pd.Series, str  ("Solve")
        Temperature in Kelvin to perform calculations at.
        Only needed for T-sensitive barometers.
        If T="Solve", returns a partial function.
        Else, enter an integer, float, or panda series.

    eq_tests: bool
        If False, just returns temperature in K (default) as a panda series.
        If True, returns pressure in kbar, Kd Fe-Mg for opx-cpx,
        and the user-entered cpx and opx comps as a panda dataframe.


    Returns
    -------
    If eq_tests is False
        pandas.Series: Pressure in kbar
    If eq_tests is True
        pandas.DataFrame: Pressure in kbar + Kd-Fe-Mg + cpx+opx comps

    '''
    try:
        func = Cpx_Opx_P_funcs_by_name[equationP]
    except KeyError:
        raise ValueError(f'{equationP} is not a valid equation') from None
    sig=inspect.signature(func)

    if sig.parameters['T'].default is not None:
        if T is None:
            raise ValueError(f'{equationP} requires you to enter T, or specify T="Solve"')
    else:
        if T is not None:
            print('Youve selected a T-independent function')

    if isinstance(T, pd.Series):
        if cpx_comps is not None:
            if len(T) != len(cpx_comps):
                raise ValueError('The panda series entered for temperature isnt the'
                ' same length as the dataframe of Cpx compositions')



    if Two_Px_Match is not None:
        two_pyx = Two_Px_Match
    if cpx_comps is not None:
        cpx_components = calculate_clinopyroxene_components(cpx_comps=cpx_comps)
        opx_components = calculate_orthopyroxene_components(opx_comps=opx_comps)
        two_pyx = pd.concat([cpx_components, opx_components], axis=1)



    kwargs = {name: two_pyx[name] for name, p in sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY}
    if isinstance(T, str) or T is None:
        if T == "Solve":
            P_kbar = partial(func, **kwargs)
        if T is None:
            P_kbar=func(**kwargs)

    else:
        P_kbar=func(T, **kwargs)

    if eq_tests is False:
        return P_kbar
    else:
        two_pyx = calculate_cpx_opx_eq_tests(
            cpx_comps=cpx_comps, opx_comps=opx_comps)
        two_pyx.insert(0, "P_kbar_calc", P_kbar)
        two_pyx.insert(2, "Equation Choice (P)", str(equationP))
        two_pyx.replace([np.inf, -np.inf], np.nan, inplace=True)
        return two_pyx



#--------------------Function for solving for temperature for two pyroxenes-----------------------------------------------------#
Cpx_Opx_T_funcs = {T_Put2008_eq36, T_Brey1990, T_Put2008_eq37, T_Wood1973, T_Wells1977} # put on outside

Cpx_Opx_T_funcs_by_name = {p.__name__: p for p in Cpx_Opx_T_funcs}


def calculate_cpx_opx_temp(*, cpx_comps=None, opx_comps=None,
                           Two_Px_Match=None, equationT=None, P=None, eq_tests=False):
    '''
    calculates temperature in Kelvin for Opx-Cpx pairs.

    The function requires inputs of cpx_comps and opx_comps, or input of a
    combined dataframe of cpx-opx compositions (this is used for the
    calculate_cpx_opx_press_temp_matching function).

    Parameters
    ------------

    cpx_comps: pandas.DataFrame
        Cpx compositions with column headings SiO2_Cpx, MgO_Cpx etc.

    opx_comps: pandas.DataFrame
       Opx compositions with column headings SiO2_Opx, MgO_Opx etc.

    Two_Px_Match: pandas.DataFrame
        Combined Cpx-Opx compositions. Used for "melt match" functionality.

    equationT: str
        Choose from:

        |  T_Put2008_Eq36  (P-dependent)
        |  T_Put2008_Eq37 (P-dependent)
        |  T_Brey1990 (P-dependent)
        |  T_Wood1973 (P-independent)
        |  T_Wells1977 (P-independent)

    P: int, float, pandas.Series, str ("Solve")
        Pressure in kbar to perform calculations at.
        Can enter float or int to use same P for all calculations
        If "Solve", returns partial if function is P-dependent

    eq_tests: bool
        If False, just returns pressure in kbar (default) as a panda series
        If True, returns pressure in kbar, Kd Fe-Mg for opx-cpx, and the user-entered cpx and opx comps as a panda dataframe.


    Returns
    -------
    If eq_tests is False
        pandas.Series: Temperature in K
    If eq_tests is True
        pandas.DataFrame: Temperature in K + Kd-Fe-Mg + cpx + opx comps

    '''

    try:
        func = Cpx_Opx_T_funcs_by_name[equationT]
    except KeyError:
        raise ValueError(f'{equationT} is not a valid equation') from None
    sig=inspect.signature(func)

    if sig.parameters['P'].default is not None:
        if P is None:
            raise ValueError(f'{equationT} requires you to enter P, or specify P="Solve"')
    else:
        if P is not None:
            print('Youve selected a P-independent function')

    if isinstance(P, pd.Series):
        if cpx_comps is not None:
            if len(P) != len(cpx_comps):
                raise ValueError('The panda series entered for Pressure isnt the same length as the dataframe of cpx compositions')



    if Two_Px_Match is not None:
        two_pyx = Two_Px_Match
    if cpx_comps is not None:
        cpx_components = calculate_clinopyroxene_components(cpx_comps=cpx_comps)
        opx_components = calculate_orthopyroxene_components(opx_comps=opx_comps)
        two_pyx = pd.concat([cpx_components, opx_components], axis=1)

    kwargs = {name: two_pyx[name] for name, p in sig.parameters.items()
    if p.kind == inspect.Parameter.KEYWORD_ONLY}


    if isinstance(P, str) or P is None:
        if P == "Solve":
            T_K = partial(func, **kwargs)
        if P is None:
            T_K=func(**kwargs)

    else:
        T_K=func(P, **kwargs)


    if eq_tests is False:
        if isinstance(T_K, partial):
            return T_K
        else:
            T_K_is_bad = (T_K == 0) | (T_K == 273.15) | (T_K ==  -np.inf) | (T_K ==  np.inf)
            T_K[T_K_is_bad] = np.nan
            return T_K

    else:
        two_pyx = calculate_cpx_opx_eq_tests(
            cpx_comps=cpx_comps, opx_comps=opx_comps)
        two_pyx.insert(0, "T_K_calc", T_K)
        two_pyx.insert(1, "Equation Choice (T)", str(equationT))

    return two_pyx

## Iterative calculations of P and T


def calculate_cpx_opx_press_temp(*, cpx_comps=None, opx_comps=None, Two_Px_Match=None,
                              equationP=None, equationT=None, iterations=30, T_K_guess=1300, eq_tests=False):
    '''
    Solves simultaneous equations for temperature and pressure using clinopyroxene-orthopyroxene
    thermometers and barometers.

    The function requires inputs of cpx_comps and opx_comps, or input of a
    combined dataframe of cpx-opx compositions (this is used for the
    calculate_cpx_opx_press_temp_matching function).

    Parameters
    -----------

    opx_comps: pandas.DataFrame
        Orthopyroxene compositions with column headings SiO2_Opx, MgO_Opx etc.

    cpx_comps: pandas.DataFrame
        Cpx compositions with column headings SiO2_Cpx, MgO_Cpx etc.

    meltmatch: pandas.DataFrame
        Combined dataframe of Opx-Cpx compositions (headings SiO2_Cpx, SiO2_Opx etc.).
        Used for calculate cpx_opx_press_temp_matching function.

    equationP: str
        Choose from:

        |  P_Put2008_eq38 (T-independent)
        |  P_Put2008_eq39 (T-dependent)

    equationT: str
        Choose from:

        |  T_Put2008_Eq36  (P-dependent)
        |  T_Put2008_Eq37 (P-dependent)
        |  T_Brey1990 (P-dependent)
        |  T_Wood1973 (P-independent)
        |  T_Wells1977 (P-independent)

    Optional:

    iterations: int, Default = 20
         Number of iterations used to converge to solution

    T_K_guess: int or float. Default = 1300K
         Initial guess of temperature to start iterations at

    eq_tests: bool
        If False, just returns pressure (default) as a panda series
        If True, returns pressure, Values of Eq tests,
        as well as user-entered opx and cpx comps and components.



    Returns
    -------
    If eq_tests is False
        pandas.DataFrame: Temperature in Kelvin, pressure in kbar
    If eq_tests is True
        pandas.DataFrame: Temperature in Kelvin, pressure in kbar, eq Tests + opx+cpx comps + components

    '''
    # Gives users flexibility to reduce or increase iterations



    if Two_Px_Match is None:
        T_func = calculate_cpx_opx_temp(
            cpx_comps=cpx_comps, opx_comps=opx_comps, equationT=equationT, P="Solve")
        P_func = calculate_cpx_opx_press(
            cpx_comps=cpx_comps, opx_comps=opx_comps, equationP=equationP, T="Solve")
    if Two_Px_Match is not None:
        T_func = calculate_cpx_opx_temp(
            Two_Px_Match=Two_Px_Match, equationT=equationT, P="Solve")
        P_func = calculate_cpx_opx_press(
            Two_Px_Match=Two_Px_Match, equationP=equationP, T="Solve")

    # This bit checks if temperature is already a series - e.g., equations
    # with no pressure dependence
    if isinstance(T_func, pd.Series) and isinstance(P_func, pd.Series):
        P_guess = P_func
        T_K_guess = T_func

    if isinstance(T_func, pd.Series) and isinstance(P_func, partial):
        P_guess = P_func(T_func)
        T_K_guess = T_func

    if isinstance(P_func, pd.Series) and isinstance(T_func, partial):
        T_K_guess = T_func(P_func)
        P_guess = P_func

    if isinstance(P_func, partial) and isinstance(T_func, partial):

        # Gives users flexibility to add a different guess temperature

        for _ in range(iterations):
            P_guess = P_func(T_K_guess)
            T_K_guess = T_func(P_guess)

        T_K_guess_is_bad = (T_K_guess == 0) | (T_K_guess == 273.15) | (T_K_guess ==  -np.inf) | (T_K_guess ==  np.inf)
        T_K_guess[T_K_guess_is_bad] = np.nan
        P_guess[T_K_guess_is_bad] = np.nan


    if eq_tests is False:
        PT_out = pd.DataFrame(data={'P_kbar_calc': P_guess,
                                    'T_K_calc': T_K_guess})

        return PT_out
    if eq_tests is True:
        two_pyx = calculate_cpx_opx_eq_tests(
            cpx_comps=cpx_comps, opx_comps=opx_comps)
        DeltaKd_HighT= np.abs(
            1.09 - two_pyx['Kd_Fe_Mg_Cpx_Opx'])
        DeltaKd_SubSol= np.abs(
            0.7 - two_pyx['Kd_Fe_Mg_Cpx_Opx'])
        two_pyx.insert(0, "T_K_calc", T_K_guess)
        two_pyx.insert(1, "P_kbar_calc", P_guess)
        two_pyx.insert(2, "Equation Choice (T)", str(equationT))
        two_pyx.insert(3, "Equation Choice (P)", str(equationP))
        two_pyx.insert(4, 'Delta Kd High T', DeltaKd_HighT)
        two_pyx.insert(6, 'Delta Kd Low T', DeltaKd_SubSol)
        two_pyx.replace([np.inf, -np.inf], np.nan, inplace=True)
    return two_pyx
   # return P_func

## Two pyroxene matching

def calculate_cpx_opx_press_temp_matching(*, opx_comps, cpx_comps, equationT=None, equationP=None,
                                  Kd_Match=None, Kd_Err=None, Cpx_Quality=False, Opx_Quality=False, P=None, T=None):
    '''
    Evaluates all possible Cpx-Opx pairs for user supplied dataframes of opx and cpx
    comps (can be different lengths). Returns P (kbar) and T (K) for those in Kd Fe-Mg equilibrium.

    Parameters
    -----------

    opx_comps: pandas.DataFrame
        Panda DataFrame of opx compositions with column headings SiO2_Opx etc.

    cpx_comps: pandas.DataFrame
        Panda DataFrame of cpx compositions with column headings SiO2_Cpx etc.

    equationP: str
        Choose from:

        |  P_Put2008_eq38 (T-independent)
        |  P_Put2008_eq39 (T-dependent)

    equationT: str
        Choose from:

        |  T_Put2008_Eq36  (P-dependent)
        |  T_Put2008_Eq37 (P-dependent)
        |  T_Brey1990 (P-dependent)
        |  T_Wood1973 (P-independent)
        |  T_Wells1977 (P-independent)

    P, T: float, int, pandas.Series. Instead of specifying equationP or equationT

        P is pressure in kbar to perform calculations at if equationP not specified

        T is temperature in Kelvin to perform calculations at if equationT not specified

    Kd_Match: str
        |  If None, returns all cpx-opx pairs.
        |  If "HighTemp", returns all cpxs-opxs within Kd cpx-opx=1.09+-0.14 suggested by Putirka (2008)
        |  If "Subsolidus" returns all cpxs-opxs within Kd cpx-opx=0.7+-0.2 suggested by Putirka (2008)
        |  If int or float, also need to specify Kd_Err. Returns all matches within Kd_Match +- Kd_Err

    Kd_Err: float or int (defaults given in Kd_Match)
        Optional input to change defaults. Returns all cpx-opx pairs within Kd_Match+-Kd_Err

    Cpx_Quality: bool
        Default False, no filter
        If True, filters out clinopyroxenes with cation sums outside of 4.02-3.99 (after Neave et al. 2017)

    Opx_Quality: bool
        Default False, no filter
        If True, filters out orthopyroxenes with cation sums outside of 4.02-3.99

    Returns
    ----------
        Dict with keys: 'Av_PTs_per_Cpx', 'All_PTs'.

        'Av_PTs_perCpx': Average P and T for each cpx, e.g., if cpx1 matches Opx1,
        Opx4, Opx6, Opx10, returns mean and 1 sigma for each cpx.

        'All_PTs': Returns output parameters for all matches, e.g, cpx1-opx1,
        cpx1-opx4 without any averaging.

    '''
    if (Kd_Err is None and isinstance(Kd_Match, int)) or (Kd_Err is None and isinstance(Kd_Match, float)):
        raise Exception(
            'You have entered a numerical value for Kd_Match, but have not'
            'specified a Kd_Err to accept matches within Kd_Match+-Kd_Err')

    # calculating Cpx and opx components. Do before duplication to save
    # computation time
    myCPXs1_concat = calculate_clinopyroxene_components(cpx_comps=cpx_comps)
    myOPXs1_concat = calculate_orthopyroxene_components(opx_comps=opx_comps)

    # Adding an ID label to help with melt-cpx rematching later
    myCPXs1_concat['ID_CPX'] = myCPXs1_concat.index
    myOPXs1_concat['ID_OPX'] = myOPXs1_concat.index
    if "Sample_ID_Cpx" not in myCPXs1_concat:
        myCPXs1_concat['Sample_ID_Cpx'] = myCPXs1_concat.index
    if "Sample_ID_Opx" not in myOPXs1_concat:
        myOPXs1_concat['Sample_ID_Opx'] = myOPXs1_concat.index
    # Duplicate cpxs and opxs so end up with panda of all possible opx-cpx
    # matches

    # This duplicates CPXs, repeats cpx1-cpx1*N, cpx2-cpx2*N etc.
    DupCPXs = pd.DataFrame(np.repeat(myCPXs1_concat.values, np.shape(
        myOPXs1_concat)[0], axis=0))  # .astype('float64')
    DupCPXs.columns = myCPXs1_concat.columns

    # This duplicates opxs like opx1-opx2-opx3 for cpx1, opx1-opx2-opx3 for
    # cpx2 etc.
    DupOPXs = pd.concat([myOPXs1_concat] * np.shape(myCPXs1_concat)[0]).reset_index(drop=True)
    # Combines these merged opx and cpx dataframes
    Combo_opxs_cpxs = pd.concat([DupOPXs, DupCPXs], axis=1)

    Combo_opxs_cpxs_1 = Combo_opxs_cpxs.copy()
    LenCombo = str(np.shape(Combo_opxs_cpxs)[0])
    print("Considering " + LenCombo +
          " Opx-Cpx pairs, be patient if this is >>1 million!")

    # calculate Kd for these pairs
    En = (Combo_opxs_cpxs.Fm2Si2O6 * (Combo_opxs_cpxs.Mg_Opx_cat_6ox /
    (Combo_opxs_cpxs.Mg_Opx_cat_6ox +Combo_opxs_cpxs.Fet_Cpx_cat_6ox + Combo_opxs_cpxs.Mn_Cpx_cat_6ox)))
    Combo_opxs_cpxs['Kd_Fe_Mg_Cpx_Opx'] = ((Combo_opxs_cpxs['Fet_Cpx_cat_6ox']
    / Combo_opxs_cpxs['Mg_Cpx_cat_6ox'])) / (Combo_opxs_cpxs['Fet_Opx_cat_6ox']
     / Combo_opxs_cpxs['Mg_Opx_cat_6ox'])

    if Kd_Match == "Subsolidus":
        Combo_opxs_cpxs['Delta_Kd_Fe_Mg_Cpx_Opx'] = np.abs(
            0.7 - Combo_opxs_cpxs['Kd_Fe_Mg_Cpx_Opx'])
        Combo_opxs_cpxs_1 = Combo_opxs_cpxs.loc[np.abs(
            Combo_opxs_cpxs['Delta_Kd_Fe_Mg_Cpx_Opx']) < 0.2]  # +- 0.2 suggested by Putirka spreadsheet
    if Kd_Match == "HighTemp":
        Combo_opxs_cpxs['Delta_Kd_Fe_Mg_Cpx_Opx'] = np.abs(
            1.09 - Combo_opxs_cpxs['Kd_Fe_Mg_Cpx_Opx'])
        # +- 0.14 suggested by Putirka spreadsheet
        Combo_opxs_cpxs_1 = Combo_opxs_cpxs.loc[np.abs(
            Combo_opxs_cpxs['Delta_Kd_Fe_Mg_Cpx_Opx']) < 0.14]
    if isinstance(Kd_Match, int) or isinstance(
            Kd_Match, float) and Kd_Err is not None:
        Combo_opxs_cpxs['Delta_Kd_Fe_Mg_Cpx_Opx'] = np.abs(
            Kd_Match - Combo_opxs_cpxs['Kd_Fe_Mg_Cpx_Opx'])
        Combo_opxs_cpxs_1 = Combo_opxs_cpxs.loc[np.abs(
            Combo_opxs_cpxs['Delta_Kd_Fe_Mg_Cpx_Opx']) < Kd_Err]  # +- 0.14 suggested by Putirka

    if Kd_Match is None and Kd_Match is None and Kd_Err is None:
        Combo_opxs_cpxs_1 = Combo_opxs_cpxs.copy()
        print('No Kd selected, all matches are shown')


    if len(Combo_opxs_cpxs_1) == 0:
        raise Exception('No matches found to the choosen Kd criteria.')

    Combo_opxs_cpxs_2 = Combo_opxs_cpxs_1.copy()
    # Uses neave method, filters out cation summs not between 4.02 and 3.99
    if Cpx_Quality is True and Opx_Quality is False:
        Combo_opxs_cpxs_2 = Combo_opxs_cpxs_1.loc[(Combo_opxs_cpxs_1['Cation_Sum_Cpx'] < 4.02) & (
            Combo_opxs_cpxs_1['Cation_Sum_Cpx'] > 3.99)]

    if Cpx_Quality is False and Opx_Quality is True:
        Combo_opxs_cpxs_2 = Combo_opxs_cpxs_1.loc[(Combo_opxs_cpxs_1['Cation_Sum_Opx'] < 4.02) & (
            Combo_opxs_cpxs_1['Cation_Sum_Opx'] > 3.99)]

    if Cpx_Quality is True and Opx_Quality is True:
        Combo_opxs_cpxs_2 = Combo_opxs_cpxs1.loc[(Combo_opxs_cpxs_1['Cation_Sum_Opx'] < 4.02) & (Combo_opxs_cpxs_1['Cation_Sum_Opx'] > 3.99)
                                                 & (Combo_opxs_cpxs_1['Cation_Sum_Cpx'] < 4.02) & (Combo_opxs_cpxs_1['Cation_Sum_Cpx'] > 3.99)]

    # Combo_opxs_cpxs_names= Combo_opxs_cpxs.copy() # Making a copy to keep
    # the names

    if len(Combo_opxs_cpxs_2) == 0:
        raise Exception('No matches found after Kd and quality filter.')


    Combo_opxs_cpxs_2_names = Combo_opxs_cpxs_2.copy()
    Combo_opxs_cpxs_2 = Combo_opxs_cpxs_2.drop(
        ['Sample_ID_Cpx', 'Sample_ID_Opx'], axis=1).astype('float64')

    # This gives users flexibility to input a constant P or T as well as
    # choosing equations
    if equationP is not None and P is not None:
        raise ValueError('You have entered an equation for P and specified a pressure. '
        ' The code doesnt know what you want it to do. Either enter an equation, or choose a pressure. ')
    if equationT is not None and T is not None:
        raise ValueError('You have entered an equation for T and specified a temperature. '
        'The code doesnt know what you want it to do. Either enter an equation, or choose a temperature.')
    if equationP is not None and equationT is not None:
        PT_out = calculate_cpx_opx_press_temp(
            Two_Px_Match=Combo_opxs_cpxs_2, equationP=equationP, equationT=equationT)
        Combo_opxs_cpxs_2.insert(0, "T_K_calc", PT_out['T_K_calc'])
        Combo_opxs_cpxs_2.insert(1, "P_kbar_calc", PT_out['P_kbar_calc'])

        Combo_opxs_cpxs_2.insert(2, "Equation Choice (T)", str(equationT))
        Combo_opxs_cpxs_2.insert(3, "Equation Choice (P)", str(equationP))
    if P is not None:
        T_K_calc = calculate_cpx_opx_temp(
            Two_Px_Match=Combo_opxs_cpxs_2, equationT=equationT, P=P)
        Combo_opxs_cpxs_2.insert(0, "T_K_calc", T_K_calc)
        Combo_opxs_cpxs_2.insert(1, "P_kbar_input", P)
        Combo_opxs_cpxs_2.insert(2, "Equation Choice (T)", str(equationT))
    if T is not None:
        P_kbar_calc = calculate_cpx_opx_press(
            Two_Px_Match=Combo_opxs_cpxs_2, equationP=equationP, T=T)
        Combo_opxs_cpxs_2.insert(0, "T_K_input", T)
        Combo_opxs_cpxs_2.insert(1, "P_kbar_calc", P_kbar_calc)
        Combo_opxs_cpxs_2.insert(3, "Equation Choice (P)", str(equationP))

    cols_to_move = ['Kd_Fe_Mg_Cpx_Opx']
    Combo_opxs_cpxs_2 = Combo_opxs_cpxs_2[cols_to_move + [
        col for col in Combo_opxs_cpxs_2.columns if col not in cols_to_move]]

    Combo_opxs_cpxs_2.insert(
        0, "Sample_ID_Opx", Combo_opxs_cpxs_2_names['Sample_ID_Opx'])
    Combo_opxs_cpxs_2.insert(
        1, "Sample_ID_Cpx", Combo_opxs_cpxs_2_names['Sample_ID_Cpx'])


        # Final step, calcuate a 3rd output which is the average and standard
        # deviation for each CPx (e.g., CPx1-Melt1, CPx1-melt3 etc. )
    CpxNumbers = Combo_opxs_cpxs_2['ID_CPX'].unique()
    if len(CpxNumbers) > 0:
        df1_Mean_nopref=Combo_opxs_cpxs_2.groupby(['ID_CPX', 'Sample_ID_Cpx'], as_index=False).mean()
        df1_Std_nopref=Combo_opxs_cpxs_2.groupby(['ID_CPX', 'Sample_ID_Cpx'], as_index=False).std()
        count=Combo_opxs_cpxs_2.groupby('ID_CPX').count()
        Sample_ID_Cpx_Mean=df1_Mean_nopref['Sample_ID_Cpx']
        Sample_ID_Cpx_Std=df1_Std_nopref['Sample_ID_Cpx']
        df1_Mean=df1_Mean_nopref.add_prefix('Mean_')
        df1_Std=df1_Std_nopref.add_prefix('Std_')
        df1_Mean=df1_Mean.drop(['Mean_Sample_ID_Cpx'], axis=1)
        df1_Std=df1_Std.drop(['Std_Sample_ID_Cpx'], axis=1)
        df1_Mean.rename(columns={"Mean_ID_CPX": "ID_CPX"}, inplace=True)
        df1_Std.rename(columns={"Std_ID_CPX": "ID_CPX"}, inplace=True)

        df1_M=pd.merge(df1_Mean, df1_Std, on=['ID_CPX'])
        df1_M['Sample_ID_Cpx']=Sample_ID_Cpx_Mean
        if equationT is not None and equationP is not None:
            cols_to_move = ['Sample_ID_Cpx',
                        'Mean_T_K_calc', 'Std_T_K_calc', 'Mean_P_kbar_calc',
                        'Std_P_kbar_calc']
        if equationT is not None and equationP is None:
            cols_to_move = ['Sample_ID_Cpx',
                        'Mean_T_K_calc', 'Std_T_K_calc', 'Mean_P_kbar_input',
                        'Std_P_kbar_input']
        if equationT is None and equationP is not None:
            cols_to_move = ['Sample_ID_Cpx',
                        'Mean_T_K_input', 'Std_T_K_input', 'Mean_P_kbar_calc',
                        'Std_P_kbar_calc']

        df1_M = df1_M[cols_to_move +
                    [col for col in df1_M.columns if col not in cols_to_move]]

    opxNumbers = Combo_opxs_cpxs_2['ID_OPX'].unique()
    if len(opxNumbers) > 0:
        df1_2Mean_nopref=Combo_opxs_cpxs_2.groupby(['ID_OPX', 'Sample_ID_Opx'], as_index=False).mean()
        df1_2Std_nopref=Combo_opxs_cpxs_2.groupby(['ID_OPX', 'Sample_ID_Opx'], as_index=False).std()
        count=Combo_opxs_cpxs_2.groupby('ID_OPX').count()
        Sample_ID_Opx_Mean=df1_2Mean_nopref['Sample_ID_Opx']
        Sample_ID_Opx_Std=df1_2Std_nopref['Sample_ID_Opx']
        df1_2Mean=df1_2Mean_nopref.add_prefix('Mean_')
        df1_2Std=df1_2Std_nopref.add_prefix('Std_')
        df1_2Mean=df1_2Mean.drop(['Mean_Sample_ID_Opx'], axis=1)
        df1_2Std=df1_2Std.drop(['Std_Sample_ID_Opx'], axis=1)
        df1_2Mean.rename(columns={"Mean_ID_OPX": "ID_OPX"}, inplace=True)
        df1_2Std.rename(columns={"Std_ID_OPX": "ID_OPX"}, inplace=True)

        df1_2M=pd.merge(df1_2Mean, df1_2Std, on=['ID_OPX'])
        df1_2M['Sample_ID_Opx']=Sample_ID_Opx_Mean


        if equationT is not None and equationP is not None:
            cols_to_move = ['Sample_ID_Opx',
                        'Mean_T_K_calc', 'Std_T_K_calc', 'Mean_P_kbar_calc',
                        'Std_P_kbar_calc']
        if equationT is not None and equationP is None:
            cols_to_move = ['Sample_ID_Opx',
                        'Mean_T_K_calc', 'Std_T_K_calc', 'Mean_P_kbar_input',
                        'Std_P_kbar_input']
        if equationT is None and equationP is not None:
            cols_to_move = ['Sample_ID_Opx',
                        'Mean_T_K_input', 'Std_T_K_input', 'Mean_P_kbar_calc',
                        'Std_P_kbar_calc']



        df1_2M = df1_2M[cols_to_move +
                    [col for col in df1_2M.columns if col not in cols_to_move]]

    else:
        raise Exception(
            'No Matches - you may need to set less strict filters, e.g.,'
            'you could edit Kd_Match is None and Kd_Err to get more matches')




    print('Done!')
    return {'Av_PTs_perCPX': df1_M, 'Av_PTs_perOPX': df1_2M, 'All_PTs': Combo_opxs_cpxs_2}






import numpy as np
import matplotlib.pyplot as plt
from functools import partial
import inspect
import warnings as w
import numbers
import pandas as pd


from Thermobar.core import *
from Thermobar.liquid_thermometers import*

## Opx-Liquid barometers



def P_Put2008_eq29a(T, *, Si_Liq_cat_frac, Mg_Liq_cat_frac, Fet_Opx_cat_6ox, FmAl2SiO6,
                    Na_Liq_cat_frac, Al_Liq_cat_frac, K_Liq_cat_frac, H2O_Liq, NaAlSi2O6):
    '''
    Orthopyroxene-Liquid barometer of Putirka, (2008) eq 29a. Global calibration of experiments.
    :cite:`putirka2008thermometers`

    SEE=+-2.6 kbar (all data)
    SEE=+-2.1 kbar for hydrous data

    '''
    Na_Si_Al_Na=(NaAlSi2O6 / (Si_Liq_cat_frac**2 * Al_Liq_cat_frac * Na_Liq_cat_frac)).astype(float)
    log_Na_Si_Al_Na=np.log(Na_Si_Al_Na)
    return (-13.97 + 0.0129 * (T - 273.15) - 19.64 * Si_Liq_cat_frac + 47.49 * Mg_Liq_cat_frac + 6.99 * Fet_Opx_cat_6ox
            + 37.37 * FmAl2SiO6 + 0.748 * H2O_Liq + 79.67 * (Na_Liq_cat_frac + K_Liq_cat_frac) +
            0.001416 * (T - 273.15)*log_Na_Si_Al_Na)



def P_Put2008_eq29b(T, *, ln_FmAl2SiO6_liq, Al_Liq_cat_frac, Mg_Liq_cat_frac,
Fet_Liq_cat_frac, Si_Opx_cat_6ox, Fet_Opx_cat_6ox,
Na_Liq_cat_frac, K_Liq_cat_frac, H2O_Liq):
    '''
    Orthopyroxene-Liquid barometer of Putirka, (2008) eq 29b. Global calibration of experiments.
    :cite:`putirka2008thermometers`

    Exact SEE not given, but ~2-3 kbar.

    '''
    return (1.788 + 0.0375 * (T - 273.15) + 0.001295 * (T - 273.15) * ln_FmAl2SiO6_liq - 33.42 * Al_Liq_cat_frac
            + 9.795 * Mg_Liq_cat_frac /
            (Mg_Liq_cat_frac + Fet_Liq_cat_frac) - 26.2 *
            Si_Opx_cat_6ox + 14.21 * Fet_Opx_cat_6ox
            + 36.08 * (Na_Liq_cat_frac + K_Liq_cat_frac) + 0.784 * H2O_Liq)


def P_Put_Global_Opx(T=None, *, MgO_Liq, Al2O3_Opx, Al2O3_Liq, Na2O_Liq, K2O_Liq):
    '''
    New Opx-Liquid barometer released in Putirka spreadsheets.
    Addresses problem in low pressure Opxs that Al(VI)=0. Uses the Al2O3 content of the Opx instead.
    :cite:`putirka2008thermometers`

    SEE=+-3.2 kbar

    '''
    return ((-8.51 + 0.856 * MgO_Liq - 1.14 * Al2O3_Opx + 45.474 *
            Al2O3_Opx / Al2O3_Liq + 1.067 * (Na2O_Liq + K2O_Liq)))


def P_Put_Felsic_Opx(T=None, *, Al2O3_Opx, Al2O3_Liq):
    '''
    New Opx_Liq barometer released in Putirka spreadsheets.
    Addresses problem in low pressure Opxs that Al(VI)=0. Uses the Al2O3 content of the Opx instead.
    Felsic regression.
    :cite:`putirka2008thermometers`

    |  SEE=+-1.2 kbar

    '''
    return ((-0.892 + 31.81 * Al2O3_Opx / Al2O3_Liq))



## Opx-Only Barometers

def P_Put2008_eq29c(T, *, Al_Opx_cat_6ox,
                    Ca_Opx_cat_6ox, Cr_Opx_cat_6ox):
    '''
    Orthopyroxene-only barometer of Putirka, (2008) eq 29c. Doesn't require liquid composition.
    Global calibration of experiments, has systematic error for hydrous data.
    :cite:`putirka2008thermometers`

    SEE=+-3 kbar (anhydrous)

    SEE=+-4.1 kbar (hydrous)

    '''
    logCr2O3 = np.log(Cr_Opx_cat_6ox.astype(float))


    return (2064 + 0.321 * (T - 273.15) - 343.4 * np.log((T - 273.15)) + 31.52 * Al_Opx_cat_6ox - 12.28 * Ca_Opx_cat_6ox
            - 290 * Cr_Opx_cat_6ox - 177.2 *
            (Al_Opx_cat_6ox - 0.1715)**2 - 372 *
            (Al_Opx_cat_6ox - 0.1715) * (Ca_Opx_cat_6ox - 0.0736)
            + 1.54 * logCr2O3)

## Opx-Liquid thermometers



def T_Put2008_eq28a(P, *, H2O_Liq, ln_Fm2Si2O6_liq, Mg_Liq_cat_frac,
                    K_Liq_cat_frac, Fet_Liq_cat_frac, Fet_Opx_cat_6ox):
    """
    Putirka (2008) Equation 28a.
    Global calibration: T=750-1600°C, SiO2=33-77 wt%, P=atm-11 GPa. H2O=0-14.2 wt%.
    :cite:`putirka2008thermometers`

    SEE= ±26°C for calibration data

    SEE=± 41°C for testing data
    """
    return (273.15 + 10**4 / (4.07 - 0.329 * (0.1 * P) + 0.12 * H2O_Liq +
    0.567 * ln_Fm2Si2O6_liq.astype(float) - 3.06 * Mg_Liq_cat_frac -
    6.17 * K_Liq_cat_frac + 1.89 * Mg_Liq_cat_frac /
    (Mg_Liq_cat_frac + Fet_Liq_cat_frac) + 2.57 * Fet_Opx_cat_6ox))

def T_Put2008_eq28b_opx_sat(P, *, H2O_Liq, Mg_Liq_cat_frac, Ca_Liq_cat_frac, K_Liq_cat_frac, Mn_Liq_cat_frac,
                            Fet_Liq_cat_frac, Fet_Opx_cat_6ox, Al_Liq_cat_frac, Ti_Liq_cat_frac, Mg_Number_Liq_NoFe3):
    '''
    Equation 28b of Putirka et al. (2008). Orthopyroxene-liquid thermometer- temperature at which a liquid is saturated in orhopyroxene (for a given P).
    :cite:`putirka2008thermometers`
    '''
    Cl_NM = Mg_Liq_cat_frac + Fet_Liq_cat_frac + \
        Ca_Liq_cat_frac + Mn_Liq_cat_frac
    NF = (7 / 2) * np.log(1 - Al_Liq_cat_frac.astype(float)) + \
        7 * np.log(1 - Ti_Liq_cat_frac.astype(float))
    return (273.15 + (5573.8 + 587.9 * (P / 10) - 61 * (P / 10)**2) / (5.3 - 0.633 * np.log(Mg_Number_Liq_NoFe3.astype(float)) - 3.97 * Cl_NM +
            0.06 * NF + 24.7 * Ca_Liq_cat_frac**2 + 0.081 * H2O_Liq + 0.156 * (P / 10)))


def T_Beatt1993_opx(P, *, Ca_Liq_cat_frac, Fet_Liq_cat_frac, Mg_Liq_cat_frac,
                    Mn_Liq_cat_frac, Al_Liq_cat_frac, Ti_Liq_cat_frac):
    '''
    Opx-Liquid thermometer of Beattie (1993). Only uses liquid composition.
    Putirka (2008) warn that overpredicts for hydrous compositions at <1200°C, and anhydrous compositions at <1100°C
    :cite:`beattie1993olivine`
    '''
    Num_B1993 = 125.9 * 1000 / 8.3144 + \
        ((0.1 * P) * 10**9 - 10**5) * 6.5 * (10**(-6)) / 8.3144
    D_Mg_opx_li1 = (0.5 - (-0.089 * Ca_Liq_cat_frac - 0.025 * Mn_Liq_cat_frac + 0.129 * Fet_Liq_cat_frac)) / \
        (Mg_Liq_cat_frac + 0.072 * Ca_Liq_cat_frac +
         0.352 * Mn_Liq_cat_frac + 0.264 * Fet_Liq_cat_frac)
    Cl_NM = Mg_Liq_cat_frac + Fet_Liq_cat_frac + \
        Ca_Liq_cat_frac + Mn_Liq_cat_frac
    NF = (7 / 2) * np.log(1 - Al_Liq_cat_frac.astype(float)) + \
        7 * np.log(1 - Ti_Liq_cat_frac.astype(float))
    Den_B1993 = 67.92 / 8.3144 + 2 * \
        np.log(D_Mg_opx_li1.astype(float)) + 2 * np.log(2 * Cl_NM.astype(float)) - NF
    return Num_B1993 / Den_B1993

##  Opx-Only barometry function
Opx_only_P_funcs = {P_Put2008_eq29c} # put on outside

Opx_only_P_funcs_by_name = {p.__name__: p for p in Opx_only_P_funcs}


def calculate_opx_only_press(*, opx_comps, equationP, T=None):
    '''
    Orthopyroxene only barometry. Enter a panda dataframe with orthopyroxene compositions, returns a pressure in kbar.

   Parameters
    -------

    opx_comps: pandas.DataFrame
        orthopyroxene compositions with column headings SiO2_Opx, MgO_Opx etc.

    equationP: str
        | P_Put2008_eq29c

    T: float, int, pandas.Series, str  ("Solve")
        Temperature in Kelvin
        Only needed for T-sensitive barometers.
        If enter T="Solve", returns a partial function
        Else, enter an integer, float, or panda series

    Returns
    -------
    pandas series
       Pressure in kbar


    '''
    try:
        func = Opx_only_P_funcs_by_name[equationP]
    except KeyError:
        raise ValueError(f'{equationP} is not a valid equation') from None

    sig=inspect.signature(func)

    if sig.parameters['T'].default is not None:
        if T is None:
            raise ValueError(f'{equationP} requires you to enter T, or specify T="Solve"')
    else:
        if T is not None:
            print('Youve selected a T-independent function')

    opx_comps = calculate_orthopyroxene_components(opx_comps=opx_comps)
    if equationP != "P_Put2008_eq29c":
        raise ValueError('Equation not recognised, at the moment the only choice is P_Put2008_eq29c')


    P_func = P_Put2008_eq29c
    if any(opx_comps['Cr_Opx_cat_6ox'] == 0):
        w.warn('The selected barometer uses the log of Cr2O3 component of '
        'Opx, which is zero for some of your compositions. '
         'This means the function will return infinity.')

    kwargs = {name: opx_comps[name] for name, p in inspect.signature(
        P_func).parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY}

    if isinstance(T, str) or T is None:
        if T == "Solve" or T is None:
            P_kbar = partial(P_func, **kwargs)

        if T == "input":
            T = liq_comps['T_K']
            P_kbar = P_func(T, **kwargs)
    else:
        T = T
        P_kbar = P_func(T, **kwargs)

    P_kbar.replace([np.inf, -np.inf], np.nan,inplace=True)



    return P_kbar

## Orthopyroxene-Liquid pressure

Opx_Liq_P_funcs = {P_Put2008_eq29a, P_Put2008_eq29b, P_Put_Global_Opx, P_Put_Felsic_Opx} # put on outside

Opx_Liq_P_funcs_by_name = {p.__name__: p for p in Opx_Liq_P_funcs}

def calculate_opx_liq_press(*, equationP, opx_comps=None, liq_comps=None, meltmatch=None,
                            T=None, eq_tests=False, H2O_Liq=None, Fe3Fet_Liq=None):
    '''
    Orthopyroxene-Liquid barometer, user specifies equation, and calculates pressure in kbar.
    Also has option to calculate equilibrium tests.

    Parameters
    -------

    opx_comps: pandas.DataFrame
        Orthopyroxene compositions with column headings SiO2_Opx, MgO_Opx etc.

    liq_comps: pandas.DataFrame
        Liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    Or:

    meltmatch: pandas.DataFrame
        Combined Opx-Liquid compositions.
        Used for calculate_opx_liq_press_temp_matching.

    EquationP: str

        choose from:

        |  P_Put2008_eq28a
        |  P_Put2008_eq28b
        |  P_Put2008_eq28c
        |  P_Put_Global_Opx
        |  P_Put_Felsic_Opx


    T: float, int, pandas.Series, str  ("Solve")
        Temperature in Kelvin
        Only needed for T-sensitive barometers.
        If enter T="Solve", returns a partial function
        Else, enter an integer, float, or panda series

    eq_tests: bool
        If False, just returns pressure (default) as a panda series
        If True, returns pressure, Values of Eq tests,
        as well as user-entered opx and liq comps and components.

    Returns
    -------
    If eq_tests=False
        pandas.Series: Pressure in kbar (if eq_tests=False)
    If eq_tests=True
        pandas.DataFrame: Pressure in kbar + Kd-Fe-Mg + opx+liq comp

    '''
# This checks if your equation is one of the accepted equations
    try:
        func = Opx_Liq_P_funcs_by_name[equationP]
    except KeyError:
        raise ValueError(f'{equationP} is not a valid equation') from None
    sig=inspect.signature(func)

    if sig.parameters['T'].default is not None:
        if T is None:
            raise ValueError(f'{equationP} requires you to enter T, or specify T="Solve"')
    else:
        if T is not None:
            print('Youve selected a T-independent function')

    if isinstance(T, pd.Series):
        if liq_comps is not None:
            if len(T) != len(liq_comps):
                raise ValueError('The panda series entered for Temperature isnt the same length as the dataframe of liquid compositions')




# This replaces H2O and Fe3Fet_Liq in the input
    if liq_comps is not None:
        liq_comps_c = liq_comps.copy()
        if H2O_Liq is not None and not isinstance(H2O_Liq, str):
            liq_comps_c['H2O_Liq'] = H2O_Liq
        if Fe3Fet_Liq is not None:
            liq_comps_c['Fe3Fet_Liq'] = Fe3Fet_Liq



    if meltmatch is not None:
        Combo_liq_opxs = meltmatch
    if liq_comps is not None:
        Combo_liq_opxs = calculate_orthopyroxene_liquid_components(
            liq_comps=liq_comps_c, opx_comps=opx_comps)

    kwargs = {name: Combo_liq_opxs[name] for name, p in sig.parameters.items() \
    if p.kind == inspect.Parameter.KEYWORD_ONLY}



    if sig.parameters['T'].default is not None:
        if T is None:
            raise ValueError(f'{equationP} requires you to enter T')
    else:
        if T is not None:
            print('Youve selected a T-independent function')


    kwargs = {name: Combo_liq_opxs[name] for name, p in sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY}
    if isinstance(T, str) or T is None:
        if T == "Solve":
            P_kbar = partial(func, **kwargs)
        if T is None:
            P_kbar=func(**kwargs)

    else:
        P_kbar=func(T, **kwargs)


    if eq_tests is False:
        if isinstance(P_kbar, partial):
            return P_kbar
        else:
            P_kbar.replace([np.inf, -np.inf], np.nan,inplace=True)
            return P_kbar

    if eq_tests is True:
        P_kbar.replace([np.inf, -np.inf], np.nan, inplace=True)
        Combo_liq_opxs.insert(1, "P_kbar_calc", P_kbar)
        Combo_liq_opxs.insert(2, "eq_tests_Kd_Fe_Mg_Fet",
                              Combo_liq_opxs['Kd_Fe_Mg_Fet'])
        Combo_liq_opxs.insert(3, "eq_tests_Kd_Fe_Mg_Fe2",
                              Combo_liq_opxs['Kd_Fe_Mg_Fe2'])
        Combo_liq_opxs.replace([np.inf, -np.inf], np.nan, inplace=True)

        return Combo_liq_opxs


## Opx-Liquid temperature

Opx_Liq_T_funcs = {T_Put2008_eq28a, T_Put2008_eq28b_opx_sat, T_Beatt1993_opx}

Opx_Liq_T_funcs_by_name = {p.__name__: p for p in Opx_Liq_T_funcs}
def calculate_opx_liq_temp(*, equationT, opx_comps=None, liq_comps=None, meltmatch=None,
                           P=None, eq_tests=False, Fe3Fet_Liq=None, H2O_Liq=None):
    '''
    Orthopyroxene-Liquid thermometer, user specifies equation,
    and calculates temperature in Kelvin.  Also has option to calculate equilibrium tests.

    Parameters
    -------

    opx_comps: pandas.DataFrame
        Orthopyroxene compositions with column headings SiO2_Opx, MgO_Opx etc.

    liq_comps: pandas.DataFrame
        Liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    meltmatch: pandas.DataFrame
        Combined Opx-Liquid compositions. Used for "melt match" functionality.

    EquationT: str
        Choice of equation:
        |  T_Opx_Beatt1993
        |  T_Put2008_eq28a
        |  T_Put2008_eq28b_opx_sat


    P: float, int, pandas.Series, str  ("Solve")
        Pressure in kbar
        Only needed for P-sensitive thermometers.
        If enter P="Solve", returns a partial function
        Else, enter an integer, float, or panda series

    eq_tests: bool
        If False (default), returns temperature as a panda series
        If True, returns prsesure, Kd Fe-Mg for liq-opx,
        as well as user-entered opx and liq comps as a panda dataframe.

    Returns
    -------
    If eq_tests=False
        pandas.Series: Pressure in kbar (if eq_tests=False)
    If eq_tests=True
        pandas.DataFrame: Pressure in kbar + Kd-Fe-Mg + opx+liq comps


    '''
    try:
        func = Opx_Liq_T_funcs_by_name[equationT]
    except KeyError:
        raise ValueError(f'{equationT} is not a valid equation') from None
    sig=inspect.signature(func)

    if sig.parameters['P'].default is not None:
        if P is None:
            raise ValueError(f'{equationT} requires you to enter P, or specify P="Solve"')
    else:
        if P is not None:
            print('Youve selected a P-independent function')

    if isinstance(P, pd.Series):
        if liq_comps is not None:
            if len(P) != len(liq_comps):
                raise ValueError('The panda series entered for Pressure isnt the same length as the dataframe of liquid compositions')



    if meltmatch is not None:
        Combo_liq_opxs = meltmatch

    if liq_comps is not None:
        liq_comps_c = liq_comps.copy()
        if H2O_Liq is not None and not isinstance(H2O_Liq, str):
            liq_comps_c['H2O_Liq'] = H2O_Liq
        if Fe3Fet_Liq is not None:
            liq_comps_c['Fe3Fet_Liq'] = Fe3Fet_Liq
        Combo_liq_opxs = calculate_orthopyroxene_liquid_components(
            liq_comps=liq_comps_c, opx_comps=opx_comps)


    kwargs = {name: Combo_liq_opxs[name] for name, p in sig.parameters.items()
    if p.kind == inspect.Parameter.KEYWORD_ONLY}


    if isinstance(P, str) or P is None:
        if P == "Solve":
            T_K = partial(func, **kwargs)
        if P is None:
            T_K=func(**kwargs)

    else:
        T_K=func(P, **kwargs)


    if eq_tests is False:
        if isinstance(T_K, partial):
            return T_K
        else:
            T_K_is_bad = (T_K == 0) | (T_K == 273.15) | (T_K ==  -np.inf) | (T_K ==  np.inf)
            T_K[T_K_is_bad] = np.nan
            return T_K


    if eq_tests is True:

        Combo_liq_opxs.insert(0, "T_K_calc", T_K)
        Combo_liq_opxs.insert(1, "eq_tests_Kd_Fe_Mg_Fet",
                              Combo_liq_opxs['Kd_Fe_Mg_Fet'])
        Combo_liq_opxs.insert(2, "eq_tests_Kd_Fe_Mg_Fe2",
                              Combo_liq_opxs['Kd_Fe_Mg_Fe2'])
        Combo_liq_opxs.replace([np.inf, -np.inf], np.nan, inplace=True)

        return Combo_liq_opxs

    return T_K

## Iterating P and T when you don't know either
def calculate_opx_liq_press_temp(*, liq_comps=None, opx_comps=None, meltmatch=None, equationP=None, equationT=None,
                              iterations=30, T_K_Guess=1300, eq_tests=False, H2O_Liq=None, Fe3Fet_Liq=None):
    '''
    Solves simultaneous equations for temperature and pressure using
    orthopyroxene-liquid thermometers and barometers.

    Parameters
    -------

    opx_comps: pandas.DataFrame
        Orthopyroxene compositions with column headings SiO2_Opx, MgO_Opx etc.

    liq_comps: pandas.DataFrame
        Liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    EquationP: str
        Barometer
        |  P_Put2008_eq28a
        |  P_Put2008_eq28b
        |  P_Put2008_eq28c
        |  P_Put_Global
        |  P_Put_Felsic

    EquationT: str
        Thermometer
        |  T_Opx_Beatt1993
        |  T_Put2008_eq28a
        |  T_Put2008_eq28b_opx_sat

    Optional:


    iterations: int (default=30)
         Number of iterations used to converge to solution.

    T_K_guess: int or float  (default=1300K)
         Initial guess of temperature. Default is 1300K

    eq_tests: bool
        If False, just returns pressure in Kbar, temp in Kelvin as a dataframe
        If True, returns pressure, temperature, Values of Eq tests,
        as well as user-entered opx and liq comps and components.


    Returns
    -------
    If eq_tests=False
        pandas.DataFrame: Temperature in Kelvin, pressure in Kbar
    If eq_tests=True
        pandas.DataFrame: Temperature in Kelvin, pressure in Kbar
        Eq Tests + opx+liq comps + components

    '''
    # Gives users flexibility to reduce or increase iterations
    if iterations is not None:
        iterations = iterations
    else:
        iterations = 30

    if T_K_Guess is not None:
        T_K_guess = T_K_Guess
    else:
        T_K_guess = 1300

    if meltmatch is None:
        liq_comps_c = liq_comps.copy()

        if Fe3Fet_Liq is not None:
            liq_comps_c['Fe3Fet_Liq'] = Fe3Fet_Liq

        if H2O_Liq is not None:
            liq_comps_c['H2O_Liq'] = H2O_Liq

        T_func = calculate_opx_liq_temp(
            opx_comps=opx_comps, liq_comps=liq_comps_c, equationT=equationT, P="Solve")
        P_func = calculate_opx_liq_press(
            opx_comps=opx_comps, liq_comps=liq_comps_c, equationP=equationP, T="Solve")

    if meltmatch is not None:

        T_func = calculate_opx_liq_temp(
            meltmatch=meltmatch, equationT=equationT, P="Solve")
        P_func = calculate_opx_liq_press(
            meltmatch=meltmatch, equationP=equationP, T="Solve")

        # Gives users flexibility to add a different guess temperature

    if isinstance(P_func, pd.Series) and isinstance(T_func, partial):
        P_guess = P_func
        T_K_guess = T_func(P_guess)
    if isinstance(T_func, pd.Series) and isinstance(P_func, partial):
        T_K_guess = T_func
        P_guess = P_func(T_K_guess)
    if isinstance(T_func, pd.Series) and isinstance(P_func, pd.Series):
        T_K_guess = T_func
        P_gues = P_func
    if isinstance(P_func, partial) and isinstance(T_func, partial):

        for _ in range(iterations):
            P_guess = P_func(T_K_guess)
            T_K_guess = T_func(P_guess)

# This gets rid of any stray Nans, 0s, 0C etc.
    T_K_guess_is_bad = (T_K_guess == 0) | (T_K_guess == 273.15) | (T_K_guess ==  -np.inf) | (T_K_guess ==  np.inf)
    T_K_guess[T_K_guess_is_bad] = np.nan
    P_guess[T_K_guess_is_bad] = np.nan

    # calculates Kd Fe-Mg if eq_tests="True"
    if eq_tests is False:
        PT_out = pd.DataFrame(
            data={'P_kbar_calc': P_guess, 'T_K_calc': T_K_guess})

        return PT_out
    if eq_tests is True and meltmatch is None:
        Combo_liq_opxs = calculate_orthopyroxene_liquid_components(
            opx_comps=opx_comps, liq_comps=liq_comps_c)
        Combo_liq_opxs.insert(0, "P_kbar_calc", P_guess)
        Combo_liq_opxs.insert(1, "T_K_calc", T_K_guess)
        Combo_liq_opxs.insert(3, "eq_tests_Kd_Fe_Mg_Fet",
                              Combo_liq_opxs['Kd_Fe_Mg_Fet'])
        Combo_liq_opxs.insert(4, "eq_tests_Kd_Fe_Mg_Fe2",
                              Combo_liq_opxs['Kd_Fe_Mg_Fe2'])
    if eq_tests is True and meltmatch is not None:
        Combo_liq_opxs = meltmatch.copy()
        Combo_liq_opxs.insert(0, "P_kbar_calc", P_guess)
        Combo_liq_opxs.insert(1, "T_K_calc", T_K_guess)
        Combo_liq_opxs.insert(3, "eq_tests_Kd_Fe_Mg_Fet",
                              meltmatch['Kd_Fe_Mg_Fet'])
        Combo_liq_opxs.insert(4, "eq_tests_Kd_Fe_Mg_Fe2",
                              meltmatch['Kd_Fe_Mg_Fe2'])

    return Combo_liq_opxs

## Considering all possible Orthopyroxene-melt pairs, and iterating P and T

def calculate_opx_liq_press_temp_matching(*, liq_comps, opx_comps, equationT=None,
equationP=None, P=None, T=None, eq_crit=False, Fe3Fet_Liq=None, H2O_Liq=None,
 Kd_Match=None, Kd_Err=None, Opx_Quality=False, return_all_pairs=False):

    '''
    Evaluates all possible Opx-Liq pairs from  N Liquids, M opx compositions
    returns P (kbar) and T (K) for those in equilibrium.

    Parameters
    -----------

    liq_comps: pandas.DataFrame
        Panda DataFrame of liquid compositions with column headings SiO2_Liq etc.

    opx_comps: pandas.DataFrame
        Panda DataFrame of opx compositions with column headings SiO2_Opx etc.

    EquationP: str
        Barometer
        |  P_Put2008_eq28a
        |  P_Put2008_eq28b
        |  P_Put2008_eq28c
        |  P_Put_Global
        |  P_Put_Felsic

    EquationT: str
        Thermometer
        |  T_Opx_Beatt1993
        |  T_Put2008_eq28a
        |  T_Put2008_eq28b_opx_sat

    Or:

    P: int, float
        Can also specify a pressure to run calculations at, rather than iterating
        using an equation for pressure. E.g., specify an equationT, but no equationP

    T: int, float
        Can also specify a temperature to run calculations at, rather than iterating
        using an equation for temperature.  E.g., specify an equationP, but no equationT

    Optional:


    Kd_Match: int of float, optional
        Allows users to ovewrite the default where Kd is calculated from the
        expression in Putirka (2008) based on the Si content of the liquid.

    Kd_Err: int or float, optional
        Allows users to override the defualt 1 sigma on Kd matches of +-0.06


    Opx Quality: bool, optional
        If True, filters out orthopyroxenes with cation sums outside of 4.02-3.99 (after Neave et al. 2017)

   Fe3Fet_Liq: int or float, optional
        Fe3FeT ratio used to assess Kd Fe-Mg equilibrium between opx and melt.
        If users don't specify, uses Fe3Fet_Liq from liq_comps.
        If specified, overwrites the Fe3Fet_Liq column in the liquid input.

    Returns: dict

        Av_PTs: Average P and T for each opx.
        E.g., if opx1 matches Liq1, Liq4, Liq6, Liq10, averages outputs for all 4 of those liquids.
        Returns mean and 1 sigma of these averaged parameters for each Opx.

        All_PTs: Returns output parameters for all matches (e.g, opx1-Liq1, opx1-Liq4) without any averaging.

    '''
    # This checks that inputs are consistent, and not contradictory
    if equationP is not None and P is not None:
        raise ValueError('You have entered an equation for P and specified a pressure. '
        ' The code doesnt know what you want it to do. Either enter an equation, or choose a pressure. ')
    if equationT is not None and T is not None:
        raise ValueError('You have entered an equation for T and specified a temperature. '
        'The code doesnt know what you want it to do. Either enter an equation, or choose a temperature.  ')

    # This over-writes inputted Fe3Fet_Liq and H2O_Liq inputs.
    liq_comps_c = liq_comps.copy()
    if Fe3Fet_Liq is not None:
        liq_comps_c['Fe3Fet_Liq'] = Fe3Fet_Liq
    if H2O_Liq is not None and not isinstance(H2O_Liq, str):
        liq_comps_c['H2O_Liq'] = H2O_Liq
    if "Fe3Fet_Liq" not in liq_comps:
        liq_comps_c['Fe3Fet_Liq'] = 0

    # Adding sample names if there aren't any
    if "Sample_ID_Liq" not in liq_comps:
        liq_comps_c['Sample_ID_Liq'] = liq_comps_c.index
    if "Sample_ID_Opx" not in opx_comps:
        opx_comps['Sample_ID_Opx'] = opx_comps.index




    # calculating Opx and liq components. Do before duplication to save
    # computation time
    myOPXs1_concat = calculate_orthopyroxene_components(opx_comps=opx_comps)
    myLiquids1_concat = calculate_anhydrous_cat_fractions_liquid(
        liq_comps=liq_comps_c)

    # Adding an ID label to help with melt-opx rematching later
    myOPXs1_concat['Sample_ID_Opx'] = opx_comps['Sample_ID_Opx']
    myLiquids1_concat['Sample_ID_Liq'] = liq_comps_c['Sample_ID_Liq']
    myOPXs1_concat['ID_OPX'] = myOPXs1_concat.index
    myLiquids1_concat['ID_Liq'] = myLiquids1_concat.index


    # This duplicates OPXs, repeats opx1-opx1*N, opx2-opx2*N etc.
    DupOPXs = pd.DataFrame(
        np.repeat(myOPXs1_concat.values, np.shape(myLiquids1_concat)[0], axis=0))
    DupOPXs.columns = myOPXs1_concat.columns

    # This duplicates liquids like liq1-liq2-liq3 for opx1, liq1-liq2-liq3 for
    # opx2 etc.
    DupLiqs = pd.concat([myLiquids1_concat] *
                        np.shape(myOPXs1_concat)[0]).reset_index(drop=True)
    # Combines these merged liquids and opx dataframes
    Combo_liq_opxs = pd.concat([DupLiqs, DupOPXs], axis=1)

    # calculate clinopyroxene-liquid components for this merged dataframe
    Combo_liq_opxs = calculate_orthopyroxene_liquid_components(
        meltmatch=Combo_liq_opxs)
   # Combo_liq_opxs.drop(['Kd Eq (Put2008+-0.06)'], axis=1, inplace=True)

    #Combo_liq_opxs = Combo_liq_opxs.convert_objects(convert_numeric=True)
    LenCombo = str(np.shape(Combo_liq_opxs)[0])
    print("Considering " + LenCombo +
          " Liq-Opx pairs, be patient if this is >>1 million!")

    if return_all_pairs is False:


        # Filters using the method of Neave et al. 2017
        if Opx_Quality is True:
            Combo_liq_opxs_2 = Combo_liq_opxs.loc[(Combo_liq_opxs['Cation_Sum_Opx'] < 4.02) & (
                Combo_liq_opxs['Cation_Sum_Opx'] > 3.99)]
        if Opx_Quality is False:
            Combo_liq_opxs_2 = Combo_liq_opxs

        # Filtering out matches which don't fit default, or user-specified Kd_Match
        # and Kd_Err values.
        if Kd_Match is None and Kd_Err is None:
            Combo_liq_opx_fur_filt = Combo_liq_opxs_2.loc[Combo_liq_opxs['Delta_Kd_Fe_Mg_Fe2'] < 0.06].reset_index(drop=True)
            Kd = Combo_liq_opx_fur_filt['Delta_Kd_Fe_Mg_Fe2']


        if Kd_Match is not None and Kd_Err is None:
            Combo_liq_opx_fur_filt = Combo_liq_opxs_2.loc[abs(Kd_Match -
            Combo_liq_opxs['Kd_Fe_Mg_Fe2']) < 0.06].reset_index(drop=True)
            Kd = Kd_Match - Combo_liq_opx_fur_filt['Kd_Fe_Mg_Fe2']
        if Kd_Match is not None and Kd_Err is not None:
            Combo_liq_opx_fur_filt = Combo_liq_opxs_2.loc[abs(Kd_Match -
            Combo_liq_opxs['Kd_Fe_Mg_Fe2']) < Kd_Err].reset_index(drop=True)
            Kd = Kd_Match - Combo_liq_opx_fur_filt['Kd_Fe_Mg_Fe2']
        if Kd_Match is None and Kd_Err is not None:
            Combo_liq_opx_fur_filt = Combo_liq_opxs_2.loc[Combo_liq_opxs['Delta_Kd_Fe_Mg_Fe2'] < Kd_Err].reset_index(drop=True)
            Kd = Combo_liq_opx_fur_filt['Delta_Kd_Fe_Mg_Fe2']

        if len(Combo_liq_opx_fur_filt) == 0:
            raise Exception('No matches found to the choosen Kd criteria.')



        # Replace automatically calculated one with various user-options.
        Combo_liq_opx_fur_filt.drop(['Delta_Kd_Fe_Mg_Fe2'], axis=1, inplace=True)
        Combo_liq_opx_fur_filt.insert(0, "Delta_Kd_Fe_Mg_Fe2", Kd)


    if return_all_pairs is True:
        Combo_liq_opx_fur_filt=Combo_liq_opxs

        # Now we have reduced down the number of calculations, we solve for P and T iteratively

        # If users want to melt match specifying an equation for both T and P
    if equationP is not None and equationT is not None:

        PT_out = calculate_opx_liq_press_temp(meltmatch=Combo_liq_opx_fur_filt, equationP=equationP, equationT=equationT)
        #print(PT_out)
        P_guess = PT_out['P_kbar_calc'].astype('float64')
        T_K_guess = PT_out['T_K_calc'].astype('float64')
        Combo_liq_opx_fur_filt.insert(0, "P_kbar_calc", P_guess.astype(float))
        Combo_liq_opx_fur_filt.insert(1, "T_K_calc", T_K_guess.astype(float))

    # Users may already know their pressure, rather than choosing an equation.
    if equationT is not None and equationP is None:
        P_guess = P
        T_K_guess = calculate_opx_liq_temp(meltmatch=Combo_liq_opx_fur_filt, equationT=equationT, P=P_guess)
        Combo_liq_opx_fur_filt.insert(0, "P_kbar_input", P_guess)
        Combo_liq_opx_fur_filt.insert(1, "T_K_calc", T_K_guess.astype(float))

# Users may already know their temperature, rather than using an equation
    if equationP is not None and equationT is None:
        T_K_guess = T
        P_guess = calculate_opx_liq_press(meltmatch=Combo_liq_opx_fur_filt, equationP=equationP, T=T_K_guess)
        Combo_liq_opx_fur_filt.insert(0, "P_kbar_calc", P_guess.astype(float))
        Combo_liq_opx_fur_filt.insert(1, "T_K_input", T_K_guess)



    print('Finished calculating Ps and Ts, now just averaging the results. Almost there!')



    # # This bit averages all the matches for a given Opx (e.g, Opx1-Liq1,
    opxNumbers = Combo_liq_opx_fur_filt['ID_OPX'].unique()
    if len(opxNumbers) > 0:
        df1_Mean_nopref=Combo_liq_opx_fur_filt.groupby(['ID_OPX', 'Sample_ID_Opx'], as_index=False).mean()
        df1_Std_nopref=Combo_liq_opx_fur_filt.groupby(['ID_OPX', 'Sample_ID_Opx'], as_index=False).std()
        count=Combo_liq_opx_fur_filt.groupby('ID_OPX').count()
        Sample_ID_Opx_Mean=df1_Mean_nopref['Sample_ID_Opx']
        Sample_ID_Opx_Std=df1_Std_nopref['Sample_ID_Opx']
        df1_Mean=df1_Mean_nopref.add_prefix('Mean_')
        df1_Std=df1_Std_nopref.add_prefix('Std_')
        df1_Mean=df1_Mean.drop(['Mean_Sample_ID_Opx'], axis=1)
        df1_Std=df1_Std.drop(['Std_Sample_ID_Opx'], axis=1)
        df1_Mean.rename(columns={"Mean_ID_OPX": "ID_OPX"}, inplace=True)
        df1_Std.rename(columns={"Std_ID_OPX": "ID_OPX"}, inplace=True)

        df1_M=pd.merge(df1_Mean, df1_Std, on=['ID_OPX'])
        df1_M['Sample_ID_Opx']=Sample_ID_Opx_Mean

        if equationT is not None and equationP is not None:
            cols_to_move = ['Sample_ID_Opx',
                        'Mean_T_K_calc', 'Std_T_K_calc', 'Mean_P_kbar_calc',
                        'Std_P_kbar_calc']

        if equationT is not None and equationP is None:
            cols_to_move = ['Sample_ID_Opx',
                        'Mean_P_kbar_input',
                        'Std_P_kbar_input', 'Mean_T_K_calc', 'Std_T_K_calc']

        if equationT is None and equationP is not None:
            cols_to_move = ['Sample_ID_Opx',
                        'Mean_T_K_input', 'Std_T_K_input', 'Mean_P_kbar_calc',
                        'Std_P_kbar_calc']



        df1_M = df1_M[cols_to_move +
                        [col for col in df1_M.columns if col not in cols_to_move]]


    else:
        raise Exception(
            'No Matches - you may need to set less strict filters, e.g.,'
            'you could edit Kd_Match is None and Kd_Err to get more matches')


    # Returns all opxs-liquids that went through 1st Kd filter with
    # equilibrium parameters, averaged matches, and all matches (not averaged)

    print('Finished!')


    cols_to_move = ['Sample_ID_Opx', 'Sample_ID_Liq']

    Combo_liq_opx_fur_filt = Combo_liq_opx_fur_filt[cols_to_move +
                        [col for col in Combo_liq_opx_fur_filt.columns if col not in cols_to_move]]

    return {'Av_PTs': df1_M, 'All_PTs': Combo_liq_opx_fur_filt}
        # return Combo_liq_opx_fur_filt